"""
FileCatalogHandler is a simple Replica and Metadata Catalog service
in the DIRAC framework
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import six
from six import StringIO
import csv
import os

from DIRAC.Core.DISET.RequestHandler import RequestHandler, getServiceOption
from DIRAC import S_OK, S_ERROR
from DIRAC.FrameworkSystem.Client.MonitoringClient import gMonitor
from DIRAC.DataManagementSystem.DB.FileCatalogDB import FileCatalogDB


class FileCatalogHandlerMixin(object):
    """
    A simple Replica and Metadata Catalog service.
    """

    @classmethod
    def initializeHandler(cls, serviceInfo):
        """Handler  initialization"""

        dbLocation = getServiceOption(serviceInfo, "Database", "DataManagement/FileCatalogDB")
        cls.fileCatalogDB = FileCatalogDB(dbLocation)

        databaseConfig = {}
        # Obtain the plugins to be used for DB interaction
        cls.log.info("Initializing with FileCatalog with following managers:")
        defaultManagers = {
            "UserGroupManager": "UserAndGroupManagerDB",
            "SEManager": "SEManagerDB",
            "SecurityManager": "NoSecurityManager",
            "DirectoryManager": "DirectoryLevelTree",
            "FileManager": "FileManager",
            "DirectoryMetadata": "DirectoryMetadata",
            "FileMetadata": "FileMetadata",
            "DatasetManager": "DatasetManager",
        }
        for configKey in sorted(defaultManagers.keys()):
            defaultValue = defaultManagers[configKey]
            configValue = getServiceOption(serviceInfo, configKey, defaultValue)
            cls.log.info("%-20s : %-20s" % (str(configKey), str(configValue)))
            databaseConfig[configKey] = configValue

        # Obtain some general configuration of the database
        cls.log.info("Initializing the FileCatalog with the following configuration:")
        defaultConfig = {
            "UniqueGUID": False,
            "GlobalReadAccess": True,
            "LFNPFNConvention": "Strong",
            "ResolvePFN": True,
            "DefaultUmask": 0o775,
            "ValidFileStatus": ["AprioriGood", "Trash", "Removing", "Probing"],
            "ValidReplicaStatus": ["AprioriGood", "Trash", "Removing", "Probing"],
            "VisibleFileStatus": ["AprioriGood"],
            "VisibleReplicaStatus": ["AprioriGood"],
        }
        for configKey in sorted(defaultConfig.keys()):
            defaultValue = defaultConfig[configKey]
            configValue = getServiceOption(serviceInfo, configKey, defaultValue)
            cls.log.info("%-20s : %-20s" % (str(configKey), str(configValue)))
            databaseConfig[configKey] = configValue
        res = cls.fileCatalogDB.setConfig(databaseConfig)

        gMonitor.registerActivity(
            "AddFile", "Amount of addFile calls", "FileCatalogHandler", "calls/min", gMonitor.OP_SUM
        )
        gMonitor.registerActivity(
            "AddFileSuccessful", "Files successfully added", "FileCatalogHandler", "files/min", gMonitor.OP_SUM
        )
        gMonitor.registerActivity(
            "AddFileFailed", "Files failed to add", "FileCatalogHandler", "files/min", gMonitor.OP_SUM
        )

        gMonitor.registerActivity(
            "RemoveFile", "Amount of removeFile calls", "FileCatalogHandler", "calls/min", gMonitor.OP_SUM
        )
        gMonitor.registerActivity(
            "RemoveFileSuccessful", "Files successfully removed", "FileCatalogHandler", "files/min", gMonitor.OP_SUM
        )
        gMonitor.registerActivity(
            "RemoveFileFailed", "Files failed to remove", "FileCatalogHandler", "files/min", gMonitor.OP_SUM
        )

        gMonitor.registerActivity(
            "AddReplica", "Amount of addReplica calls", "FileCatalogHandler", "calls/min", gMonitor.OP_SUM
        )
        gMonitor.registerActivity(
            "AddReplicaSuccessful", "Replicas successfully added", "FileCatalogHandler", "replicas/min", gMonitor.OP_SUM
        )
        gMonitor.registerActivity(
            "AddReplicaFailed", "Replicas failed to add", "FileCatalogHandler", "replicas/min", gMonitor.OP_SUM
        )

        gMonitor.registerActivity(
            "RemoveReplica", "Amount of removeReplica calls", "FileCatalogHandler", "calls/min", gMonitor.OP_SUM
        )
        gMonitor.registerActivity(
            "RemoveReplicaSuccessful",
            "Replicas successfully removed",
            "FileCatalogHandler",
            "replicas/min",
            gMonitor.OP_SUM,
        )
        gMonitor.registerActivity(
            "RemoveReplicaFailed", "Replicas failed to remove", "FileCatalogHandler", "replicas/min", gMonitor.OP_SUM
        )

        gMonitor.registerActivity(
            "ListDirectory", "Amount of listDirectory calls", "FileCatalogHandler", "calls/min", gMonitor.OP_SUM
        )

        return res

    ########################################################################
    # Path operations (not updated)
    #
    types_changePathOwner = [[list, dict] + list(six.string_types)]

    def export_changePathOwner(self, lfns, recursive=False):
        """Get replica info for the given list of LFNs"""
        return self.fileCatalogDB.changePathOwner(lfns, self.getRemoteCredentials(), recursive)

    types_changePathGroup = [[list, dict] + list(six.string_types)]

    def export_changePathGroup(self, lfns, recursive=False):
        """Get replica info for the given list of LFNs"""
        return self.fileCatalogDB.changePathGroup(lfns, self.getRemoteCredentials(), recursive)

    types_changePathMode = [[list, dict] + list(six.string_types)]

    def export_changePathMode(self, lfns, recursive=False):
        """Get replica info for the given list of LFNs"""
        return self.fileCatalogDB.changePathMode(lfns, self.getRemoteCredentials(), recursive)

    ########################################################################
    # ACL Operations
    #
    types_getPathPermissions = [[list, dict] + list(six.string_types)]

    def export_getPathPermissions(self, lfns):
        """Determine the ACL information for a supplied path"""
        return self.fileCatalogDB.getPathPermissions(lfns, self.getRemoteCredentials())

    types_hasAccess = [[six.string_types, dict], [six.string_types, list, dict]]

    def export_hasAccess(self, paths, opType):
        """Determine if the given op can be performed on the paths
        The OpType is all the operations exported

        The reason for the param types is backward compatibility. Between v6r14 and v6r15,
        the signature of hasAccess has changed, and the two parameters were swapped.
        """

        # The signature of v6r15 is (dict, str)
        # The signature of v6r14 is (str, [dict, str, list])
        # We swap the two params if the first attribute is a string
        if isinstance(paths, six.string_types):
            paths, opType = opType, paths

        return self.fileCatalogDB.hasAccess(opType, paths, self.getRemoteCredentials())

    ###################################################################
    #
    #  isOK
    #

    types_isOK = []

    @classmethod
    def export_isOK(cls):
        """returns S_OK if DB is connected"""
        if cls.fileCatalogDB and cls.fileCatalogDB._connected:
            return S_OK()
        return S_ERROR("Server not connected to DB")

    ###################################################################
    #
    #  User/Group write operations
    #

    types_addUser = [six.string_types]

    def export_addUser(self, userName):
        """Add a new user to the File Catalog"""
        return self.fileCatalogDB.addUser(userName, self.getRemoteCredentials())

    types_deleteUser = [six.string_types]

    def export_deleteUser(self, userName):
        """Delete user from the File Catalog"""
        return self.fileCatalogDB.deleteUser(userName, self.getRemoteCredentials())

    types_addGroup = [six.string_types]

    def export_addGroup(self, groupName):
        """Add a new group to the File Catalog"""
        return self.fileCatalogDB.addGroup(groupName, self.getRemoteCredentials())

    types_deleteGroup = [six.string_types]

    def export_deleteGroup(self, groupName):
        """Delete group from the File Catalog"""
        return self.fileCatalogDB.deleteGroup(groupName, self.getRemoteCredentials())

    ###################################################################
    #
    #  User/Group read operations
    #

    types_getUsers = []

    def export_getUsers(self):
        """Get all the users defined in the File Catalog"""
        return self.fileCatalogDB.getUsers(self.getRemoteCredentials())

    types_getGroups = []

    def export_getGroups(self):
        """Get all the groups defined in the File Catalog"""
        return self.fileCatalogDB.getGroups(self.getRemoteCredentials())

    ########################################################################
    #
    # Path read operations
    #

    types_exists = [[list, dict] + list(six.string_types)]

    def export_exists(self, lfns):
        """Check whether the supplied paths exists"""
        return self.fileCatalogDB.exists(lfns, self.getRemoteCredentials())

    ########################################################################
    #
    # File write operations
    #

    types_addFile = [[list, dict] + list(six.string_types)]

    def export_addFile(self, lfns):
        """Register supplied files"""
        gMonitor.addMark("AddFile", 1)
        res = self.fileCatalogDB.addFile(lfns, self.getRemoteCredentials())
        if res["OK"]:
            gMonitor.addMark("AddFileSuccessful", len(res.get("Value", {}).get("Successful", [])))
            gMonitor.addMark("AddFileFailed", len(res.get("Value", {}).get("Failed", [])))

        return res

    types_removeFile = [[list, dict] + list(six.string_types)]

    def export_removeFile(self, lfns):
        """Remove the supplied lfns"""
        gMonitor.addMark("RemoveFile", 1)
        res = self.fileCatalogDB.removeFile(lfns, self.getRemoteCredentials())
        if res["OK"]:
            gMonitor.addMark("RemoveFileSuccessful", len(res.get("Value", {}).get("Successful", [])))
            gMonitor.addMark("RemoveFileFailed", len(res.get("Value", {}).get("Failed", [])))

        return res

    types_setFileStatus = [dict]

    def export_setFileStatus(self, lfns):
        """Remove the supplied lfns"""
        return self.fileCatalogDB.setFileStatus(lfns, self.getRemoteCredentials())

    types_addReplica = [[list, dict] + list(six.string_types)]

    def export_addReplica(self, lfns):
        """Register supplied replicas"""
        gMonitor.addMark("AddReplica", 1)
        res = self.fileCatalogDB.addReplica(lfns, self.getRemoteCredentials())
        if res["OK"]:
            gMonitor.addMark("AddReplicaSuccessful", len(res.get("Value", {}).get("Successful", [])))
            gMonitor.addMark("AddReplicaFailed", len(res.get("Value", {}).get("Failed", [])))

        return res

    types_removeReplica = [[list, dict] + list(six.string_types)]

    def export_removeReplica(self, lfns):
        """Remove the supplied replicas"""
        gMonitor.addMark("RemoveReplica", 1)
        res = self.fileCatalogDB.removeReplica(lfns, self.getRemoteCredentials())
        if res["OK"]:
            gMonitor.addMark("RemoveReplicaSuccessful", len(res.get("Value", {}).get("Successful", [])))
            gMonitor.addMark("RemoveReplicaFailed", len(res.get("Value", {}).get("Failed", [])))

        return res

    types_setReplicaStatus = [[list, dict] + list(six.string_types)]

    def export_setReplicaStatus(self, lfns):
        """Set the status for the supplied replicas"""
        return self.fileCatalogDB.setReplicaStatus(lfns, self.getRemoteCredentials())

    types_setReplicaHost = [[list, dict] + list(six.string_types)]

    def export_setReplicaHost(self, lfns):
        """Change the registered SE for the supplied replicas"""
        return self.fileCatalogDB.setReplicaHost(lfns, self.getRemoteCredentials())

    types_addFileAncestors = [dict]

    def export_addFileAncestors(self, lfns):
        """Add file ancestor information for the given list of LFNs"""
        return self.fileCatalogDB.addFileAncestors(lfns, self.getRemoteCredentials())

    ########################################################################
    #
    # File read operations
    #

    types_isFile = [[list, dict] + list(six.string_types)]

    def export_isFile(self, lfns):
        """Check whether the supplied lfns are files"""
        return self.fileCatalogDB.isFile(lfns, self.getRemoteCredentials())

    types_getFileSize = [[list, dict] + list(six.string_types)]

    def export_getFileSize(self, lfns):
        """Get the size associated to supplied lfns"""
        return self.fileCatalogDB.getFileSize(lfns, self.getRemoteCredentials())

    types_getFileMetadata = [[list, dict] + list(six.string_types)]

    def export_getFileMetadata(self, lfns):
        """Get the metadata associated to supplied lfns"""
        return self.fileCatalogDB.getFileMetadata(lfns, self.getRemoteCredentials())

    types_getReplicas = [[list, dict] + list(six.string_types), bool]

    def export_getReplicas(self, lfns, allStatus=False):
        """Get replicas for supplied lfns"""
        return self.fileCatalogDB.getReplicas(lfns, allStatus, self.getRemoteCredentials())

    types_getReplicaStatus = [[list, dict] + list(six.string_types)]

    def export_getReplicaStatus(self, lfns):
        """Get the status for the supplied replicas"""
        return self.fileCatalogDB.getReplicaStatus(lfns, self.getRemoteCredentials())

    types_getFileAncestors = [[list, dict], (list,) + six.integer_types]

    def export_getFileAncestors(self, lfns, depths):
        """Get the status for the supplied replicas"""
        dList = depths
        if not isinstance(dList, list):
            dList = [depths]
        lfnDict = dict.fromkeys(lfns, True)
        return self.fileCatalogDB.getFileAncestors(lfnDict, dList, self.getRemoteCredentials())

    types_getFileDescendents = [[list, dict], (list,) + six.integer_types]

    def export_getFileDescendents(self, lfns, depths):
        """Get the status for the supplied replicas"""
        dList = depths
        if not isinstance(dList, list):
            dList = [depths]
        lfnDict = dict.fromkeys(lfns, True)
        return self.fileCatalogDB.getFileDescendents(lfnDict, dList, self.getRemoteCredentials())

    types_getLFNForGUID = [[list, dict] + list(six.string_types)]

    def export_getLFNForGUID(self, guids):
        """Get the matching lfns for given guids"""
        return self.fileCatalogDB.getLFNForGUID(guids, self.getRemoteCredentials())

    ########################################################################
    #
    # Directory write operations
    #

    types_createDirectory = [[list, dict] + list(six.string_types)]

    def export_createDirectory(self, lfns):
        """Create the supplied directories"""
        return self.fileCatalogDB.createDirectory(lfns, self.getRemoteCredentials())

    types_removeDirectory = [[list, dict] + list(six.string_types)]

    def export_removeDirectory(self, lfns):
        """Remove the supplied directories"""
        return self.fileCatalogDB.removeDirectory(lfns, self.getRemoteCredentials())

    ########################################################################
    #
    # Directory read operations
    #

    types_listDirectory = [[list, dict] + list(six.string_types), bool]

    def export_listDirectory(self, lfns, verbose):
        """List the contents of supplied directories"""
        gMonitor.addMark("ListDirectory", 1)
        return self.fileCatalogDB.listDirectory(lfns, self.getRemoteCredentials(), verbose=verbose)

    types_isDirectory = [[list, dict] + list(six.string_types)]

    def export_isDirectory(self, lfns):
        """Determine whether supplied path is a directory"""
        return self.fileCatalogDB.isDirectory(lfns, self.getRemoteCredentials())

    types_getDirectoryMetadata = [[list, dict] + list(six.string_types)]

    def export_getDirectoryMetadata(self, lfns):
        """Get the size of the supplied directory"""
        return self.fileCatalogDB.getDirectoryMetadata(lfns, self.getRemoteCredentials())

    types_getDirectorySize = [[list, dict] + list(six.string_types)]

    def export_getDirectorySize(self, lfns, longOut=False, fromFiles=False, recursiveSum=True):
        """Get the size of the supplied directory"""
        return self.fileCatalogDB.getDirectorySize(lfns, longOut, fromFiles, recursiveSum, self.getRemoteCredentials())

    types_getDirectoryReplicas = [[list, dict] + list(six.string_types), bool]

    def export_getDirectoryReplicas(self, lfns, allStatus=False):
        """Get replicas for files in the supplied directory"""
        return self.fileCatalogDB.getDirectoryReplicas(lfns, allStatus, self.getRemoteCredentials())

    ########################################################################
    #
    # Administrative database operations
    #

    types_getCatalogCounters = []

    def export_getCatalogCounters(self):
        """Get the number of registered directories, files and replicas in various tables"""
        return self.fileCatalogDB.getCatalogCounters(self.getRemoteCredentials())

    types_rebuildDirectoryUsage = []

    def export_rebuildDirectoryUsage(self):
        """Rebuild DirectoryUsage table from scratch"""
        return self.fileCatalogDB.rebuildDirectoryUsage()

    types_repairCatalog = []

    def export_repairCatalog(self):
        """Repair the catalog inconsistencies"""
        return self.fileCatalogDB.repairCatalog(self.getRemoteCredentials())

    ########################################################################
    # Metadata Catalog Operations
    #

    types_addMetadataField = [six.string_types, six.string_types, six.string_types]

    def export_addMetadataField(self, fieldName, fieldType, metaType="-d"):
        """Add a new metadata field of the given type"""
        if metaType.lower() == "-d":
            return self.fileCatalogDB.dmeta.addMetadataField(fieldName, fieldType, self.getRemoteCredentials())
        elif metaType.lower() == "-f":
            return self.fileCatalogDB.fmeta.addMetadataField(fieldName, fieldType, self.getRemoteCredentials())
        else:
            return S_ERROR("Unknown metadata type %s" % metaType)

    types_deleteMetadataField = [six.string_types]

    def export_deleteMetadataField(self, fieldName):
        """Delete the metadata field"""
        result = self.fileCatalogDB.dmeta.deleteMetadataField(fieldName, self.getRemoteCredentials())
        error = ""
        if not result["OK"]:
            error = result["Message"]
        result = self.fileCatalogDB.fmeta.deleteMetadataField(fieldName, self.getRemoteCredentials())
        if not result["OK"]:
            if error:
                result["Message"] = error + "; " + result["Message"]

        return result

    types_getMetadataFields = []

    def export_getMetadataFields(self):
        """Get all the metadata fields"""
        resultDir = self.fileCatalogDB.dmeta.getMetadataFields(self.getRemoteCredentials())
        if not resultDir["OK"]:
            return resultDir
        resultFile = self.fileCatalogDB.fmeta.getFileMetadataFields(self.getRemoteCredentials())
        if not resultFile["OK"]:
            return resultFile

        return S_OK({"DirectoryMetaFields": resultDir["Value"], "FileMetaFields": resultFile["Value"]})

    types_setMetadata = [six.string_types, dict]

    def export_setMetadata(self, path, metadatadict):
        """Set metadata parameter for the given path"""
        return self.fileCatalogDB.setMetadata(path, metadatadict, self.getRemoteCredentials())

    types_setMetadataBulk = [dict]

    def export_setMetadataBulk(self, pathMetadataDict):
        """Set metadata parameter for the given path"""
        return self.fileCatalogDB.setMetadataBulk(pathMetadataDict, self.getRemoteCredentials())

    types_removeMetadata = [dict]

    def export_removeMetadata(self, pathMetadataDict):
        """Remove the specified metadata for the given path"""
        return self.fileCatalogDB.removeMetadata(pathMetadataDict, self.getRemoteCredentials())

    types_getDirectoryUserMetadata = [six.string_types]

    def export_getDirectoryUserMetadata(self, path):
        """Get all the metadata valid for the given directory path"""
        return self.fileCatalogDB.dmeta.getDirectoryMetadata(path, self.getRemoteCredentials())

    types_getFileUserMetadata = [six.string_types]

    def export_getFileUserMetadata(self, path):
        """Get all the metadata valid for the given file"""
        return self.fileCatalogDB.fmeta.getFileUserMetadata(path, self.getRemoteCredentials())

    types_findDirectoriesByMetadata = [dict]

    def export_findDirectoriesByMetadata(self, metaDict, path="/"):
        """Find all the directories satisfying the given metadata set"""
        return self.fileCatalogDB.dmeta.findDirectoriesByMetadata(metaDict, path, self.getRemoteCredentials())

    types_findFilesByMetadata = [dict, six.string_types]

    def export_findFilesByMetadata(self, metaDict, path="/"):
        """Find all the files satisfying the given metadata set"""
        result = self.fileCatalogDB.fmeta.findFilesByMetadata(metaDict, path, self.getRemoteCredentials())
        if not result["OK"]:
            return result
        lfns = list(result["Value"].values())
        return S_OK(lfns)

    types_getReplicasByMetadata = [dict, six.string_types, bool]

    def export_getReplicasByMetadata(self, metaDict, path="/", allStatus=False):
        """Find all the files satisfying the given metadata set"""
        return self.fileCatalogDB.fileManager.getReplicasByMetadata(
            metaDict, path, allStatus, self.getRemoteCredentials()
        )

    types_findFilesByMetadataDetailed = [dict, six.string_types]

    def export_findFilesByMetadataDetailed(self, metaDict, path="/"):
        """Find all the files satisfying the given metadata set"""
        result = self.fileCatalogDB.fmeta.findFilesByMetadata(metaDict, path, self.getRemoteCredentials())
        if not result["OK"] or not result["Value"]:
            return result

        lfns = list(result["Value"].values())
        return self.fileCatalogDB.getFileDetails(lfns, self.getRemoteCredentials())

    types_findFilesByMetadataWeb = [dict, six.string_types, six.integer_types, six.integer_types]

    def export_findFilesByMetadataWeb(self, metaDict, path, startItem, maxItems):
        """Find files satisfying the given metadata set"""
        result = self.fileCatalogDB.dmeta.findFileIDsByMetadata(
            metaDict, path, self.getRemoteCredentials(), startItem, maxItems
        )
        if not result["OK"] or not result["Value"]:
            return result

        fileIDs = result["Value"]
        totalRecords = result["TotalRecords"]

        result = self.fileCatalogDB.fileManager._getFileLFNs(fileIDs)
        if not result["OK"]:
            return result

        lfnsResultList = list(result["Value"]["Successful"].values())
        resultDetails = self.fileCatalogDB.getFileDetails(lfnsResultList, self.getRemoteCredentials())
        if not resultDetails["OK"]:
            return resultDetails

        result = S_OK({"TotalRecords": totalRecords, "Records": resultDetails["Value"]})
        return result

    def findFilesByMetadataWeb(self, metaDict, path, startItem, maxItems):
        """Find all the files satisfying the given metadata set"""
        result = self.fileCatalogDB.fmeta.findFilesByMetadata(metaDict, path, self.getRemoteCredentials())
        if not result["OK"] or not result["Value"]:
            return result

        lfns = []
        for directory in result["Value"]:
            for fname in result["Value"][directory]:
                lfns.append(os.path.join(directory, fname))

        start = startItem
        totalRecords = len(lfns)
        if start > totalRecords:
            return S_ERROR("Requested files out of existing range")
        end = start + maxItems
        if end > totalRecords:
            end = totalRecords
        lfnsResultList = lfns[start:end]

        resultDetails = self.fileCatalogDB.getFileDetails(lfnsResultList, self.getRemoteCredentials())
        if not resultDetails["OK"]:
            return resultDetails

        result = S_OK({"TotalRecords": totalRecords, "Records": resultDetails["Value"]})
        return result

    types_getCompatibleMetadata = [dict, six.string_types]

    def export_getCompatibleMetadata(self, metaDict, path="/"):
        """Get metadata values compatible with the given metadata subset"""
        return self.fileCatalogDB.dmeta.getCompatibleMetadata(metaDict, path, self.getRemoteCredentials())

    types_addMetadataSet = [six.string_types, dict]

    def export_addMetadataSet(self, setName, setDict):
        """Add a new metadata set"""
        return self.fileCatalogDB.dmeta.addMetadataSet(setName, setDict, self.getRemoteCredentials())

    types_getMetadataSet = [six.string_types, bool]

    def export_getMetadataSet(self, setName, expandFlag):
        """Add a new metadata set"""
        return self.fileCatalogDB.dmeta.getMetadataSet(setName, expandFlag, self.getRemoteCredentials())

    #########################################################################################
    #
    #  Dataset manipulation methods
    #
    types_addDataset = [dict]

    def export_addDataset(self, datasets):
        """Add a new dynamic dataset defined by its meta query"""
        return self.fileCatalogDB.datasetManager.addDataset(datasets, self.getRemoteCredentials())

    types_addDatasetAnnotation = [dict]

    def export_addDatasetAnnotation(self, datasetDict):
        """Add annotation to an already created dataset"""
        return self.fileCatalogDB.datasetManager.addDatasetAnnotation(datasetDict, self.getRemoteCredentials())

    types_removeDataset = [dict]

    def export_removeDataset(self, datasets):
        """Check the given dynamic dataset for changes since its definition"""
        return self.fileCatalogDB.datasetManager.removeDataset(datasets, self.getRemoteCredentials())

    types_checkDataset = [dict]

    def export_checkDataset(self, datasets):
        """Check the given dynamic dataset for changes since its definition"""
        return self.fileCatalogDB.datasetManager.checkDataset(datasets, self.getRemoteCredentials())

    types_updateDataset = [dict]

    def export_updateDataset(self, datasets):
        """Update the given dynamic dataset for changes since its definition"""
        return self.fileCatalogDB.datasetManager.updateDataset(datasets, self.getRemoteCredentials())

    types_getDatasets = [dict]

    def export_getDatasets(self, datasets):
        """Get parameters of the given dynamic dataset as they are stored in the database"""
        return self.fileCatalogDB.datasetManager.getDatasets(datasets, self.getRemoteCredentials())

    types_getDatasetParameters = [dict]

    def export_getDatasetParameters(self, datasets):
        """Get parameters of the given dynamic dataset as they are stored in the database"""
        return self.fileCatalogDB.datasetManager.getDatasetParameters(datasets, self.getRemoteCredentials())

    types_getDatasetAnnotation = [dict]

    def export_getDatasetAnnotation(self, datasets):
        """Get annotation of the given datasets"""
        return self.fileCatalogDB.datasetManager.getDatasetAnnotation(datasets, self.getRemoteCredentials())

    types_freezeDataset = [dict]

    def export_freezeDataset(self, datasets):
        """Freeze the contents of the dataset making it effectively static"""
        return self.fileCatalogDB.datasetManager.freezeDataset(datasets, self.getRemoteCredentials())

    types_releaseDataset = [dict]

    def export_releaseDataset(self, datasets):
        """Release the contents of the frozen dataset allowing changes in its contents"""
        return self.fileCatalogDB.datasetManager.releaseDataset(datasets, self.getRemoteCredentials())

    types_getDatasetFiles = [dict]

    def export_getDatasetFiles(self, datasets):
        """Get lfns in the given dataset"""
        return self.fileCatalogDB.datasetManager.getDatasetFiles(datasets, self.getRemoteCredentials())

    def getSEDump(self, seName):
        """
         Return all the files at a given SE, together with checksum and size

        :param seName: name of the StorageElement

        :returns: S_OK with list of tuples (lfn, checksum, size)
        """
        return self.fileCatalogDB.getSEDump(seName)["Value"]


class FileCatalogHandler(FileCatalogHandlerMixin, RequestHandler):
    def transfer_toClient(self, seName, token, fileHelper):
        """This method used to transfer the SEDump to the client,
        formated as CSV with '|' separation

        :param seName: name of the se to dump

        :returns: the result of the FileHelper


        """

        retVal = self.getSEDump(seName)

        try:
            csvOutput = StringIO()
            writer = csv.writer(csvOutput, delimiter="|")
            for lfn in retVal:
                writer.writerow(lfn)

            csvOutput.seek(0)

            ret = fileHelper.DataSourceToNetwork(csvOutput)
            return ret

        except Exception as e:
            self.log.exception("Exception while sending seDump", repr(e))
            return S_ERROR("Exception while sending seDump: %s" % repr(e))
        finally:
            csvOutput.close()



"""
Message Queue wrapper
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import time
from pythonjsonlogger.jsonlogger import JsonFormatter

from DIRAC.FrameworkSystem.private.standardLogging.Handler.MessageQueueHandler import MessageQueueHandler
from DIRAC.FrameworkSystem.private.standardLogging.LogLevels import LogLevels
from DIRAC.Resources.LogBackends.AbstractBackend import AbstractBackend

DEFAULT_MQ_LEVEL = "verbose"
# These are the standard logging fields that we want to see
# in the json. All the non default are printed anyway
DEFAULT_FMT = "%(levelname)s %(message)s %(asctime)s"


class MicrosecondJsonFormatter(JsonFormatter):
    def formatTime(self, record, datefmt=None):
        """:py:meth:`logging.Formatter.formatTime` with microsecond precision by default"""
        ct = self.converter(record.created)
        if datefmt:
            s = time.strftime(datefmt, ct)
        else:
            t = time.strftime("%Y-%m-%d %H:%M:%S", ct)
            s = "%s,%06d" % (t, (record.created - int(record.created)) * 1e6)
        return s


class MessageQueueBackend(AbstractBackend):
    """
    MessageQueueBackend is used to create an abstraction of the handler and the formatter concepts from logging.
    Here, we have:

      - MessageQueueHandler: which is a custom handler created in DIRAC to send
        log records to a Message Queue server. You can find it in: FrameworkSys./private/standardlogging/Handler
      - BaseFormatter: is a custom Formatter object, created for DIRAC in order to get the appropriate display.

      You can find it in FrameworkSystem/private/standardLogging/Formatter
    """

    def __init__(self, backendParams=None):
        """
        Initialization of the MessageQueueBackend
        """
        # The `Format` parameter is passed as `fmt` to MicrosecondJsonFormatter
        # which uses it to know which "standard" fields to keep in the
        # json output. So we need these
        if not backendParams:
            backendParams = {}
        backendParams.setdefault("Format", DEFAULT_FMT)

        super(MessageQueueBackend, self).__init__(
            MessageQueueHandler, MicrosecondJsonFormatter, backendParams, level=DEFAULT_MQ_LEVEL
        )

    def _setHandlerParameters(self, backendParams=None):
        """
        Get the handler parameters from the backendParams.
        The keys of handlerParams should correspond to the parameter names of the associated handler.
        The method should be overridden in every backend that needs handler parameters.
        The method should be called before creating the handler object.

        :param dict parameters: parameters of the backend. ex: {'FileName': file.log}
        """
        # default values
        self._handlerParams["queue"] = ""

        if backendParams is not None:
            self._handlerParams["queue"] = backendParams.get("MsgQueue", self._handlerParams["queue"])



"""
Backend wrapper
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.FrameworkSystem.private.standardLogging.LogLevels import LogLevels


class AbstractBackend(object):
    """
    AbstractBackend is used to create an abstraction of handler and formatter concepts from 'logging'.
    It corresponds to the backend concept of the old gLogger.
    It is useful for different reasons:

    - to gather handler and formatter,
      in DIRAC, each handler must be bind with a specific formatter so
      it is more simple and more clear to create an object for this job.

    - each backend can get its options and
      format them to give them to the handler,
      in DIRAC, it is possible to add backend options in the cfgfile.
      For example, for a file, you can give the filename that you want to write log inside.
      The backend allows to each handler to get its own options as parameters. Again, it is more clear
      and more simple to have a specific object for this job.

    In this way, we have an object composed by one handler and one formatter name.
    The purpose of the object is to get cfg options to give them to the handler,
    and to set the format of the handler when the display must be changed.
    """

    def __init__(self, handlerType, formatterType, backendParams=None, level="debug"):
        """
        Initialization of the backend.
        _handler and _formatter can be custom objects. If it is the case, you can find them
        in FrameworkSystem/private/standardLogging/Formatter or Handler.

        :param _handler: handler object from 'logging'. Ex: StreamHandler(), FileHandler()...
        :param _formatter: the name of a formatter object from logging. Ex: BaseFormatter
        :param dict backendParams: parameters to set up the backend
        :param str _datefmt: parameters to set up the formatter (e.g. fmt, the format, and datefmt, the date format)
        :param str _level: level of the handler

        """
        # get handler parameters from the backendParams and instantiate the handler
        self._handlerParams = {}
        self._setHandlerParameters(backendParams)
        self._setHandler(handlerType)

        # get formatter parameters from the backendParams, instantiate and attach the formatter to the handler
        self._formatterParams = {}
        self._setFormatterParameters(backendParams)
        self._setFormatter(formatterType)

        # set the level: can also be defined in the backendParams
        if backendParams:
            level = backendParams.get("LogLevel", level)
        self.setLevel(level)

    def getHandler(self):
        """
        :return: the handler
        """
        return self._handler

    def _setHandler(self, handlerType):
        """
        Instantiate a handler from the given handlerType.

        :param str handlerType: the handler type (e.g. logging.StreamHandler)
        """
        self._handler = handlerType(**self._handlerParams)

    def _setHandlerParameters(self, backendParams=None):
        """
        Get the handler parameters from the backendParams.
        The keys of handlerParams should correspond to the parameter names of the associated handler.
        The method should be overridden in every backend that needs handler parameters.
        The method should be called before creating the handler object.

        :param dict parameters: parameters of the backend. ex: {'FileName': file.log}
        """
        pass

    def _setFormatter(self, formatterType):
        """
        Instantiate a formatter from the given formatterType and attach it to the handler.

        :param str formatterType: the formatter type (e.g. BaseFormatter)
        """
        self._handler.setFormatter(formatterType(**self._formatterParams))

    def _setFormatterParameters(self, backendParams=None):
        """
        Get the formatting option from the backendParams

        :param dict backendParams: parameters of the backend
        """
        # Default values
        self._formatterParams["fmt"] = None
        self._formatterParams["datefmt"] = None

        # Get values from formatterParameters
        if backendParams is not None:
            self._formatterParams["fmt"] = backendParams.get("Format")
            self._formatterParams["datefmt"] = backendParams.get("DateFormat")

    def setLevel(self, levelName):
        """
        Configure the level of the handler associated to the backend.
        Make sure the handler has been created before calling the method.

        :param int level: a level
        """
        result = False
        if levelName.upper() in LogLevels.getLevelNames():
            self._handler.setLevel(LogLevels.getLevelValue(levelName))
            self._level = levelName
            result = True
        return result



""" Utilities for pretty printing table data and more
    Author: A.Tsaregorodtsev
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import six
from six import StringIO


def int_with_commas(inputValue):
    """Utility to make a string of a large integer with comma separated degrees
        of thousand

    :param int inputValue: value to be interpreted
    :return: output string
    """
    s = str(inputValue)
    news = ""
    while len(s) > 0:
        news = s[-3:] + "," + news
        s = s[:-3]
    return news[:-1]


def printTable(fields, records, sortField="", numbering=True, printOut=True, columnSeparator=" "):
    """Utility to pretty print tabular data

    :param list fields: list of column names
    :param list records: list of records, each record is a list or tuple of field values
           where field value itself can be a string or a list of strings or a dictionary
           of the structure { "Value": string_value, 'Just': 'L|R|C' } to specify justification
    :param str sortField: name of the column by which the output will be sorted
    :param bool numbering: flag for numbering rows
    :param bool printOut: flag for printing into the stdout
    :param str columnSeparator: string to be used as a column separator
    :return: pretty table string
    """

    def __writeField(buffer, value, length, columnSeparator, lastColumn=False):

        justification = None
        if isinstance(value, dict):
            justification = value.get("Just")
            value = value.get("Value", "")

        if justification is None:
            # try casting to int and then align to the right, if it fails align to the left
            justification = "r"
            try:
                _ = int("".join(value.split(",")))
            except ValueError:
                justification = "l"

        if justification.lower() == "l":
            output = value.ljust(length) + columnSeparator
        elif justification.lower() == "r":
            output = value.rjust(length) + columnSeparator
        elif justification.lower() == "c":
            margin = length - len(value)
            if margin <= 1:
                output = value.ljust(length) + columnSeparator
            else:
                m1 = int(margin / 2)
                m2 = margin - m1
                output = " " * m1 + value + " " * m2 + columnSeparator
        if lastColumn:
            output = output.rstrip()
        buffer.write(output)
        return len(output)

    if not records:
        if printOut:
            print("No output")
        return "No output"

    # Strip all strings
    fieldList = [f.strip() for f in fields]
    recordList = []
    for record in records:
        strippedRecord = []
        for fieldValue in record:
            if isinstance(fieldValue, six.string_types):
                strippedRecord.append(fieldValue.strip())
            elif isinstance(fieldValue, list):
                strippedList = []
                for ll in fieldValue:
                    if isinstance(ll, six.string_types):
                        strippedList.append(ll.strip())
                    elif isinstance(ll, dict):
                        ll["Value"] = ll["Value"].strip()
                        strippedList.append(ll)
                    else:
                        out = "Wrong type for field value: %s" % type(ll)
                        if printOut:
                            print(out)
                        return out
                strippedRecord.append(strippedList)
            elif isinstance(fieldValue, dict):
                itemValue = fieldValue["Value"]
                if isinstance(itemValue, six.string_types):
                    itemValue = itemValue.strip()
                    fieldValue.update({"Value": itemValue})
                    strippedRecord.append(fieldValue)
                elif isinstance(itemValue, list):
                    itemValue = [r.strip() for r in itemValue]
                    fieldValue.update({"Value": itemValue})
                    strippedRecord.append(fieldValue)
                else:
                    out = "Wrong type for field value: %s" % type(itemValue)
                    if printOut:
                        print(out)
                    return out
        recordList.append(strippedRecord)

    nFields = len(fields)
    for rec in records:
        if nFields != len(rec):
            out = "Incorrect data structure to print, nFields %d, nRecords %d" % (nFields, len(rec))
            if printOut:
                print(out)
            return out

    if sortField:
        recordList.sort(key=lambda x: x[fieldList.index(sortField)])

    # Compute the maximum width for each field
    fieldWidths = []
    for i in range(nFields):
        fieldWidths.append(len(fieldList[i]))
        for record in recordList:
            if isinstance(record[i], list):
                for item in record[i]:
                    if isinstance(item, dict):
                        itemValue = item["Value"]
                    else:
                        itemValue = item
                    fieldWidths[i] = max(len(itemValue), fieldWidths[i])
            elif isinstance(record[i], dict):
                fieldValue = record[i]["Value"]
                if isinstance(fieldValue, list):
                    fieldWidths[i] = max(max(len(item) for item in fieldValue), fieldWidths[i])
                else:
                    fieldWidths[i] = max(len(fieldValue), fieldWidths[i])
            else:
                fieldWidths[i] = max(len(record[i]), fieldWidths[i])

    numberWidth = len(str(len(recordList))) + 1
    separatorWidth = len(columnSeparator)
    totalLength = sum(fieldWidths) + separatorWidth * nFields
    if numbering:
        totalLength += numberWidth + separatorWidth

    # Accumulate the table output in the stringBuffer now
    stringBuffer = StringIO()
    topLength = (numberWidth + separatorWidth) if numbering else 0
    stringBuffer.write(" " * (topLength))

    for i in range(nFields):
        lastColumn = False if i < nFields - 1 else True
        length = __writeField(stringBuffer, fieldList[i], fieldWidths[i], columnSeparator, lastColumn)
        topLength += length
    stringBuffer.write("\n")
    if columnSeparator == " ":
        stringBuffer.write("=" * topLength + "\n")
    else:
        stringBuffer.write("=" * totalLength + "\n")

    for count, record in enumerate(recordList):
        total = count == len(recordList) - 1 and recordList[-1][0] == "Total"
        if numbering:
            if total:
                # Do not number the line with the total
                stringBuffer.write(" " * (numberWidth + separatorWidth))
            else:
                stringBuffer.write(str(count + 1).rjust(numberWidth) + columnSeparator)

        listMode = 0
        for item in record:
            if isinstance(item, list):
                listMode = max(len(item), listMode)
            elif isinstance(item, dict) and isinstance(item["Value"], list):
                listMode = max(len(item["Value"]), listMode)

        for i, (fieldValue, fieldWidth) in enumerate(zip(record, fieldWidths)):
            lastColumn = False if i < nFields - 1 else True
            value = fieldValue
            if isinstance(fieldValue, list):
                value = fieldValue[0]
            elif isinstance(fieldValue, dict) and isinstance(fieldValue["Value"], list):
                value = dict(fieldValue)
                value.update({"Value": value["Value"][0]})

            __writeField(stringBuffer, value, fieldWidth, columnSeparator, lastColumn)

        # If the field has a list type value, print out one value per line
        if listMode:
            stringBuffer.write("\n")
            for ll in range(1, listMode):
                # Do not number continuation lines
                if numbering:
                    stringBuffer.write(" " * (numberWidth + separatorWidth))
                for i, (fieldValue, fieldWidth) in enumerate(zip(record, fieldWidths)):
                    lastColumn = False if i < nFields - 1 else True
                    valueList = fieldValue
                    if isinstance(valueList, list) and ll < len(valueList):
                        value = valueList[ll]
                    elif (
                        isinstance(valueList, dict)
                        and isinstance(valueList["Value"], list)
                        and ll < len(valueList["Value"])
                    ):
                        value = dict(valueList)
                        value.update({"Value": valueList["Value"][ll]})
                    else:
                        value = ""

                    __writeField(stringBuffer, value, fieldWidth, columnSeparator, lastColumn)

                stringBuffer.write("\n")

        if not listMode:
            stringBuffer.write("\n")
        if total:
            stringBuffer.write("-" * totalLength + "\n")

    output = stringBuffer.getvalue()
    if printOut:
        print(output)

    return output


def printDict(dDict, printOut=False):
    """Utility to pretty print a dictionary

    :param dict dDict: Dictionary to be printed out
    :param bool printOut: flag to print to the stdout
    :return: pretty dictionary string
    """
    lines = []
    keyLength = 0
    for key in dDict:
        if len(key) > keyLength:
            keyLength = len(key)
    for key in sorted(dDict):
        line = "%s: " % key
        line = line.ljust(keyLength + 2)
        value = dDict[key]
        if isinstance(value, (list, tuple)):
            line += ",".join(list(value))
        else:
            line += str(value)
        lines.append(line)
    output = "{\n%s\n}" % "\n".join(lines)
    if printOut:
        print(output)
    return output



""" test for RFIO plugin
"""

# FIXME: doesn't work ATM


from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
import unittest
import time
import os
import shutil
import mock

from DIRAC import S_OK

from DIRAC.Resources.Storage.test.Test_FilePlugin import (
    mock_StorageFactory_getConfigStorageOptions,
    mock_StorageFactory_getConfigStorageProtocols,
    mock_StorageFactory_getConfigStorageName,
)

# from DIRAC.Resources.Storage.StorageFactory     import StorageFactory
from DIRAC.Resources.Storage.StorageElement import StorageElementItem
from DIRAC.Core.Utilities.File import getSize

__RCSID__ = "$Id$"


def mock_StorageFactory_getCurrentURL(storageName, derivedStorageName):
    """Get the options associated to the StorageElement as defined in the CS"""
    optionsDict = {"BackendType": "local", "ReadAccess": "Active", "WriteAccess": "Active"}

    return S_OK(optionsDict)


# class StoragePlugInTestCase( unittest.TestCase ):
#   """ Base class for the StoragePlugin test cases
#   """
#   def setUp( self ):
#     factory = StorageFactory()
#     res = factory.getStorages( 'CERN-RAW', ['RFIO'] )
#     self.assertTrue(res['OK'])
#     storageDetails = res['Value']


class StoragePlugInTestCase(unittest.TestCase):
    """Base test class. Defines all the method to test"""

    @mock.patch(
        "DIRAC.Resources.Storage.StorageFactory.StorageFactory._getConfigStorageName",
        side_effect=mock_StorageFactory_getConfigStorageName,
    )
    @mock.patch(
        "DIRAC.Resources.Storage.StorageFactory.StorageFactory._getConfigStorageOptions",
        side_effect=mock_StorageFactory_getConfigStorageOptions,
    )
    @mock.patch(
        "DIRAC.Resources.Storage.StorageFactory.StorageFactory._getConfigStorageProtocols",
        side_effect=mock_StorageFactory_getConfigStorageProtocols,
    )
    @mock.patch(
        "DIRAC.Resources.Storage.StorageElement.StorageElementItem._StorageElementItem__isLocalSE",
        return_value=S_OK(True),
    )  # Pretend it's local
    @mock.patch(
        "DIRAC.Resources.Storage.StorageElement.StorageElementItem.addAccountingOperation", return_value=None
    )  # Don't send accounting
    # @mock.patch( 'DIRAC.Resources.Storage.StorageFactory.StorageFactory._getCurrentURL',
    #              side_effect = mock_StorageFactory_getCurrentURL_getCurrentURL )
    def setUp(
        self,
        mk_getConfigStorageName,
        mk_getConfigStorageOptions,
        mk_getConfigStorageProtocols,
        mk_isLocalSE,
        mk_addAccountingOperation,
    ):
        self.storage = StorageElementItem("FAKE")
        self.storage.vo = "test"

        # self.storage = storageDetails['StorageObjects'][0]
        # self.storage.changeDirectory( 'lhcb/test/unit-test/Storage/RFIOStorage' )

    def test_createUnitTestDir(self):
        print(
            "\n\n#########################################################"
            "################\n\n\t\t\tCreate Directory test\n"
        )
        # destDir = self.storage.getCurrentURL( '' )['Value']
        destDir = "/bla/"
        res = self.storage.createDirectory(destDir)
        print(res)
        self.assertTrue(res["OK"])
        self.assertTrue(destDir in res["Value"]["Successful"])
        self.assertTrue(res["Value"]["Successful"][destDir])


class DirectoryTestCase(StoragePlugInTestCase):
    def test_isDirectory(self):
        print(
            "\n\n#########################################################"
            "################\n\n\t\t\tIs Directory test\n"
        )
        # Test that we can determine what is a directory
        destDir = self.storage.getCurrentURL("")["Value"]
        isDirRes = self.storage.isDirectory(destDir)
        # Test that we can determine that a directory is not a directory
        destDir = self.storage.getCurrentURL("NonExistantFile")["Value"]
        nonExistantDirRes = self.storage.isDirectory(destDir)

        # Check the is directory operation
        self.assertTrue(isDirRes["OK"])
        self.assertTrue(destDir in isDirRes["Value"]["Successful"])
        self.assertTrue(isDirRes["Value"]["Successful"][destDir])
        # Check the non existant directory operation
        self.assertTrue(nonExistantDirRes["OK"])
        self.assertTrue(destDir in nonExistantDirRes["Value"]["Successful"])
        self.assertFalse(nonExistantDirRes["Value"]["Successful"][destDir])

    def test_putRemoveDirectory(self):
        print(
            "\n\n#########################################################"
            "################\n\n\t\t\tPut Directory test\n"
        )
        # First clean the remote directory incase something was left there
        remoteDir = self.storage.getCurrentURL("putDirTest")["Value"]
        ignore = self.storage.removeDirectory(remoteDir)

        # Create a local directory to upload
        localDir = "/tmp/unit-test"
        srcFile = "/etc/group"
        numberOfFiles = 5
        sizeOfLocalFile = getSize(srcFile)
        print("Creating local directory: %s" % localDir)
        if not os.path.exists(localDir):
            os.mkdir(localDir)
        for i in range(numberOfFiles):
            shutil.copy(srcFile, "%s/testFile.%s" % (localDir, time.time()))
            time.sleep(1)
        # Check that we can successfully upload the directory to the storage element
        dirTuple = (localDir, remoteDir)
        putDirRes = self.storage.putDirectory(dirTuple)
        # Now remove the remove directory
        removeDirRes = self.storage.removeDirectory(remoteDir)
        # Clean up the locally created directory
        print("Removing local directory: %s" % localDir)
        localFiles = os.listdir(localDir)
        for fileName in localFiles:
            fullPath = "%s/%s" % (localDir, fileName)
            os.remove(fullPath)
        os.removedirs(localDir)

        # Perform the checks for the put dir operation
        self.assertTrue(putDirRes["OK"])
        self.assertTrue(remoteDir in putDirRes["Value"]["Successful"])
        resDict = putDirRes["Value"]["Successful"][remoteDir]
        self.assertEqual(resDict["Files"], numberOfFiles)
        self.assertEqual(resDict["Size"], numberOfFiles * sizeOfLocalFile)
        # Perform the checks for the remove dir operation
        self.assertTrue(removeDirRes["OK"])
        self.assertTrue(remoteDir in removeDirRes["Value"]["Successful"])
        resDict = removeDirRes["Value"]["Successful"][remoteDir]
        self.assertEqual(resDict["Files"], numberOfFiles)
        self.assertEqual(resDict["Size"], numberOfFiles * sizeOfLocalFile)

    def test_putGetDirectoryMetadata(self):
        print(
            "\n\n#########################################################"
            "################\n\n\t\t\tGet Directory Metadata test\n"
        )
        # First clean the remote directory incase something was left there
        remoteDir = self.storage.getCurrentURL("putDirTest")["Value"]
        ignore = self.storage.removeDirectory(remoteDir)

        # Create a local directory to upload
        localDir = "/tmp/unit-test"
        srcFile = "/etc/group"
        numberOfFiles = 5
        sizeOfLocalFile = getSize(srcFile)
        print("Creating local directory: %s" % localDir)
        if not os.path.exists(localDir):
            os.mkdir(localDir)
        for i in range(numberOfFiles):
            shutil.copy(srcFile, "%s/testFile.%s" % (localDir, time.time()))
            time.sleep(1)
        # Check that we can successfully upload the directory to the storage element
        dirTuple = (localDir, remoteDir)
        putDirRes = self.storage.putDirectory(dirTuple)
        # Get the directory metadata
        getMetadataRes = self.storage.getDirectoryMetadata(remoteDir)
        # Now remove the remove directory
        removeDirRes = self.storage.removeDirectory(remoteDir)
        # Clean up the locally created directory
        print("Removing local directory: %s" % localDir)
        localFiles = os.listdir(localDir)
        for fileName in localFiles:
            fullPath = "%s/%s" % (localDir, fileName)
            os.remove(fullPath)
        os.removedirs(localDir)

        # Perform the checks for the put dir operation
        self.assertTrue(putDirRes["OK"])
        self.assertTrue(remoteDir in putDirRes["Value"]["Successful"])
        resDict = putDirRes["Value"]["Successful"][remoteDir]
        self.assertEqual(resDict["Files"], numberOfFiles)
        self.assertEqual(resDict["Size"], numberOfFiles * sizeOfLocalFile)
        # Perform the checks for the get metadata operation
        self.assertTrue(getMetadataRes["OK"])
        self.assertTrue(remoteDir in getMetadataRes["Value"]["Successful"])
        resDict = getMetadataRes["Value"]["Successful"][remoteDir]
        self.assertTrue("Mode" in resDict)
        self.assertEqual(resDict["Mode"], 493)
        # Perform the checks for the remove directory operation
        self.assertTrue(removeDirRes["OK"])
        self.assertTrue(remoteDir in removeDirRes["Value"]["Successful"])
        resDict = removeDirRes["Value"]["Successful"][remoteDir]
        self.assertEqual(resDict["Files"], numberOfFiles)
        self.assertEqual(resDict["Size"], numberOfFiles * sizeOfLocalFile)

    def test_putGetDirectorySize(self):
        print(
            "\n\n#########################################################"
            "################\n\n\t\t\tGet Directory Size test\n"
        )
        # First clean the remote directory incase something was left there
        remoteDir = self.storage.getCurrentURL("putDirTest")["Value"]
        ignore = self.storage.removeDirectory(remoteDir)

        # Create a local directory to upload
        localDir = "/tmp/unit-test"
        srcFile = "/etc/group"
        numberOfFiles = 5
        sizeOfLocalFile = getSize(srcFile)
        print("Creating local directory: %s" % localDir)
        if not os.path.exists(localDir):
            os.mkdir(localDir)
        for i in range(numberOfFiles):
            shutil.copy(srcFile, "%s/testFile.%s" % (localDir, time.time()))
            time.sleep(1)
        # Check that we can successfully upload the directory to the storage element
        dirTuple = (localDir, remoteDir)
        putDirRes = self.storage.putDirectory(dirTuple)
        # Now get the directory size
        getDirSizeRes = self.storage.getDirectorySize(remoteDir)
        # Now remove the remove directory
        removeDirRes = self.storage.removeDirectory(remoteDir)
        # Clean up the locally created directory
        print("Removing local directory: %s" % localDir)
        localFiles = os.listdir(localDir)
        for fileName in localFiles:
            fullPath = "%s/%s" % (localDir, fileName)
            os.remove(fullPath)
        os.removedirs(localDir)

        # Perform the checks for the put dir operation
        self.assertTrue(putDirRes["OK"])
        self.assertTrue(remoteDir in putDirRes["Value"]["Successful"])
        resDict = putDirRes["Value"]["Successful"][remoteDir]
        self.assertEqual(resDict["Files"], numberOfFiles)
        self.assertEqual(resDict["Size"], numberOfFiles * sizeOfLocalFile)
        # Now perform the checks for the get directory size operation
        self.assertTrue(getDirSizeRes["OK"])
        self.assertTrue(remoteDir in getDirSizeRes["Value"]["Successful"])
        resDict = getDirSizeRes["Value"]["Successful"][remoteDir]
        self.assertEqual(resDict["Size"], numberOfFiles * sizeOfLocalFile)
        self.assertEqual(resDict["Files"], numberOfFiles)
        # Perform the checks for the remove directory operation
        self.assertTrue(removeDirRes["OK"])
        self.assertTrue(remoteDir in removeDirRes["Value"]["Successful"])
        resDict = removeDirRes["Value"]["Successful"][remoteDir]
        self.assertEqual(resDict["Files"], numberOfFiles)
        self.assertEqual(resDict["Size"], numberOfFiles * sizeOfLocalFile)

    def test_putListDirectory(self):
        print(
            "\n\n#########################################################"
            "################\n\n\t\t\tList Directory test\n"
        )
        # First clean the remote directory incase something was left there
        remoteDir = self.storage.getCurrentURL("putDirTest")["Value"]
        ignore = self.storage.removeDirectory(remoteDir)

        # Create a local directory to upload
        localDir = "/tmp/unit-test"
        srcFile = "/etc/group"
        numberOfFiles = 5
        sizeOfLocalFile = getSize(srcFile)
        print("Creating local directory: %s" % localDir)
        if not os.path.exists(localDir):
            os.mkdir(localDir)
        for i in range(numberOfFiles):
            shutil.copy(srcFile, "%s/testFile.%s" % (localDir, time.time()))
            time.sleep(1)
        # Check that we can successfully upload the directory to the storage element
        dirTuple = (localDir, remoteDir)
        putDirRes = self.storage.putDirectory(dirTuple)
        # List the remote directory
        listDirRes = self.storage.listDirectory(remoteDir)
        # Now remove the remove directory
        removeDirRes = self.storage.removeDirectory(remoteDir)
        # Clean up the locally created directory
        print("Removing local directory: %s" % localDir)
        localFiles = os.listdir(localDir)
        for fileName in localFiles:
            fullPath = "%s/%s" % (localDir, fileName)
            os.remove(fullPath)
        os.removedirs(localDir)

        # Perform the checks for the put dir operation
        self.assertTrue(putDirRes["OK"])
        self.assertTrue(remoteDir in putDirRes["Value"]["Successful"])
        resDict = putDirRes["Value"]["Successful"][remoteDir]
        self.assertEqual(resDict["Files"], numberOfFiles)
        self.assertEqual(resDict["Size"], numberOfFiles * sizeOfLocalFile)
        # Perform the checks for the list dir operation
        self.assertTrue(listDirRes["OK"])
        self.assertTrue(remoteDir in listDirRes["Value"]["Successful"])
        resDict = listDirRes["Value"]["Successful"][remoteDir]
        self.assertTrue("SubDirs" in resDict)
        self.assertTrue("Files" in resDict)
        self.assertEqual(len(resDict["Files"]), numberOfFiles)
        # Perform the checks for the remove directory operation
        self.assertTrue(removeDirRes["OK"])
        self.assertTrue(remoteDir in removeDirRes["Value"]["Successful"])
        resDict = removeDirRes["Value"]["Successful"][remoteDir]
        self.assertEqual(resDict["Files"], numberOfFiles)
        self.assertEqual(resDict["Size"], numberOfFiles * sizeOfLocalFile)

    def test_putGetDirectory(self):
        print(
            "\n\n#########################################################"
            "################\n\n\t\t\tGet Directory test\n"
        )
        # First clean the remote directory incase something was left there
        remoteDir = self.storage.getCurrentURL("putDirTest")["Value"]
        ignore = self.storage.removeDirectory(remoteDir)

        # Create a local directory to upload
        localDir = "/tmp/unit-test"
        srcFile = "/etc/group"
        numberOfFiles = 5
        sizeOfLocalFile = getSize(srcFile)
        print("Creating local directory: %s" % localDir)
        if not os.path.exists(localDir):
            os.mkdir(localDir)
        for i in range(numberOfFiles):
            shutil.copy(srcFile, "%s/testFile.%s" % (localDir, time.time()))
            time.sleep(1)
        # Check that we can successfully upload the directory to the storage element
        dirTuple = (localDir, remoteDir)
        putDirRes = self.storage.putDirectory(dirTuple)
        # Clean up the locally created directory
        print("Removing local directory: %s" % localDir)
        localFiles = os.listdir(localDir)
        for fileName in localFiles:
            fullPath = "%s/%s" % (localDir, fileName)
            os.remove(fullPath)
        os.removedirs(localDir)
        # Check that we can get directories from the storage element
        directoryTuple = (remoteDir, localDir)
        getDirRes = self.storage.getDirectory(directoryTuple)
        # Now remove the remove directory
        removeDirRes = self.storage.removeDirectory(remoteDir)
        # Clean up the locally created directory
        print("Removing local directory: %s" % localDir)
        localFiles = os.listdir(localDir)
        for fileName in localFiles:
            fullPath = "%s/%s" % (localDir, fileName)
            os.remove(fullPath)
        os.removedirs(localDir)

        # Perform the checks for the put dir operation
        self.assertTrue(putDirRes["OK"])
        self.assertTrue(remoteDir in putDirRes["Value"]["Successful"])
        resDict = putDirRes["Value"]["Successful"][remoteDir]
        self.assertEqual(resDict["Files"], numberOfFiles)
        self.assertEqual(resDict["Size"], numberOfFiles * sizeOfLocalFile)
        # Perform the checks for the get dir operation
        self.assertTrue(getDirRes["OK"])
        self.assertTrue(remoteDir in getDirRes["Value"]["Successful"])
        resDict = getDirRes["Value"]["Successful"][remoteDir]
        self.assertEqual(resDict["Files"], numberOfFiles)
        self.assertEqual(resDict["Size"], numberOfFiles * sizeOfLocalFile)
        # Perform the checks for the remove directory operation
        self.assertTrue(removeDirRes["OK"])
        self.assertTrue(remoteDir in removeDirRes["Value"]["Successful"])
        resDict = removeDirRes["Value"]["Successful"][remoteDir]
        self.assertEqual(resDict["Files"], numberOfFiles)
        self.assertEqual(resDict["Size"], numberOfFiles * sizeOfLocalFile)


class FileTestCase(StoragePlugInTestCase):
    def test_putRemoveFile(self):
        print(
            "\n\n#########################################################"
            "################\n\n\t\t\tPut and Remove test\n"
        )
        # First test that we are able to determine whether the file sizes of the transfer don't match
        srcFile = "/etc/group"
        fileSize = 10  # This is a made up value
        testFileName = "testFile.%s" % time.time()
        destFile = self.storage.getCurrentURL(testFileName)["Value"]
        fileTuple = (srcFile, destFile, fileSize)
        failedPutFileRes = self.storage.putFile(fileTuple)
        # Now make sure that we can actually upload a file properly
        fileSize = getSize(srcFile)
        fileTuple = (srcFile, destFile, fileSize)
        putFileRes = self.storage.putFile(fileTuple)
        # Make sure we are able to remove the file
        removeFileRes = self.storage.removeFile(destFile)

        # Check the failed put file operation
        self.assertTrue(failedPutFileRes["OK"])
        self.assertTrue(destFile in failedPutFileRes["Value"]["Failed"])
        expectedError = "RFIOStorage.putFile: Source and destination file sizes do not match."
        self.assertEqual(failedPutFileRes["Value"]["Failed"][destFile], expectedError)
        # Check the successful put file operation
        self.assertTrue(putFileRes["OK"])
        self.assertTrue(destFile in putFileRes["Value"]["Successful"])
        # Check the remove file operation
        self.assertTrue(removeFileRes["OK"])
        self.assertTrue(destFile in removeFileRes["Value"]["Successful"])

    """
  def test_putGetFile(self):
    print '\n\n#########################################################################\n\n\t\t\tPut and Get test\n'
    # First upload a file to the storage
    srcFile = '/etc/group'
    fileSize = getSize(srcFile)
    testFileName = 'testFile.%s' % time.time()
    destFile = self.storage.getCurrentURL(testFileName)['Value']
    fileTuple = (srcFile,destFile,fileSize)
    putRes = self.storage.putFile(fileTuple)
    # Make sure we can detect when the file transferred incorrectly
    destLocalFile = "%s/%s.%s" % (os.getcwd(),'unitTestFile',time.time())
    wrongSize = 10
    fileTuple = (destFile,destLocalFile,wrongSize)
    failedGetRes = self.storage.getFile(fileTuple)
    # Then make sure we can get a local copy of the file
    destLocalFile = "%s/%s.%s" % (os.getcwd(),'unitTestFile',time.time())
    fileTuple = (destFile,destLocalFile,fileSize)
    getFileRes = self.storage.getFile(fileTuple)
    # Cleanup the remote mess
    removeFileRes = self.storage.removeFile(destFile)
    #Cleanup the mess locally
    os.remove(destLocalFile)

    # Check the put operation
    self.assertTrue(putRes['OK'])
    self.assertTrue(putRes['Value']['Successful'].has_key(destFile))
    # Check the failed get operation
    self.assertTrue(failedGetRes['OK'])
    self.assertTrue(failedGetRes['Value']['Failed'].has_key(destFile))
    expectedError = 'RFIOStorage.getFile: Source and destination file sizes do not match.'
    self.assertEqual(failedGetRes['Value']['Failed'][destFile],expectedError)
    # Check the get operation
    self.assertTrue(getFileRes['OK'])
    self.assertTrue(getFileRes['Value']['Successful'].has_key(destFile))
    # Check the remove operation
    self.assertTrue(removeFileRes['OK'])
    self.assertTrue(removeFileRes['Value']['Successful'].has_key(destFile))

  """

    def test_putExistsFile(self):
        print("\n\n#########################################################" "################\n\n\t\t\tExists test\n")
        # First upload a file to the storage
        srcFile = "/etc/group"
        fileSize = getSize(srcFile)
        testFileName = "testFile.%s" % time.time()
        destFile = self.storage.getCurrentURL(testFileName)["Value"]
        fileTuple = (srcFile, destFile, fileSize)
        putFileRes = self.storage.putFile(fileTuple)
        # Then get the file's existance
        existsFileRes = self.storage.exists(destFile)
        # Now remove the file
        removeFileRes = self.storage.removeFile(destFile)
        # Check  again that the file exists
        failedExistRes = self.storage.exists(destFile)

        # Check the put file operation
        self.assertTrue(putFileRes["OK"])
        self.assertTrue(destFile in putFileRes["Value"]["Successful"])
        # Check the exists operation
        self.assertTrue(existsFileRes["OK"])
        self.assertTrue(destFile in existsFileRes["Value"]["Successful"])
        self.assertTrue(existsFileRes["Value"]["Successful"][destFile])
        # Check the removal operation
        self.assertTrue(removeFileRes["OK"])
        self.assertTrue(destFile in removeFileRes["Value"]["Successful"])
        # Check the failed exists operation
        self.assertTrue(failedExistRes["OK"])
        self.assertTrue(destFile in failedExistRes["Value"]["Successful"])
        self.assertFalse(failedExistRes["Value"]["Successful"][destFile])

    def test_putIsFile(self):
        print(
            "\n\n#########################################################" "################\n\n\t\t\tIs file test\n"
        )
        # First upload a file to the storage
        srcFile = "/etc/group"
        fileSize = getSize(srcFile)
        testFileName = "testFile.%s" % time.time()
        destFile = self.storage.getCurrentURL(testFileName)["Value"]
        fileTuple = (srcFile, destFile, fileSize)
        putFileRes = self.storage.putFile(fileTuple)
        # Check we are able to determine that it is a file
        isFileRes = self.storage.isFile(destFile)
        # Clean up the remote mess
        removeFileRes = self.storage.removeFile(destFile)
        # Check that everything isn't a file
        destDir = os.path.dirname(destFile)
        failedIsFileRes = self.storage.isFile(destDir)

        # Check the put file operation
        self.assertTrue(putFileRes["OK"])
        self.assertTrue(destFile in putFileRes["Value"]["Successful"])
        # Check the is file operation
        self.assertTrue(isFileRes["OK"])
        self.assertTrue(destFile in isFileRes["Value"]["Successful"])
        self.assertTrue(isFileRes["Value"]["Successful"][destFile])
        # check the remove file operation
        self.assertTrue(removeFileRes["OK"])
        self.assertTrue(destFile in removeFileRes["Value"]["Successful"])
        # Check that the directory is not a file
        self.assertTrue(failedIsFileRes["OK"])
        self.assertTrue(destDir in failedIsFileRes["Value"]["Successful"])
        self.assertFalse(failedIsFileRes["Value"]["Successful"][destDir])

    def test_putGetFileMetaData(self):
        print(
            "\n\n#########################################################"
            "################\n\n\t\t\tGet file metadata test\n"
        )
        # First upload a file to the storage
        srcFile = "/etc/group"
        fileSize = getSize(srcFile)
        testFileName = "testFile.%s" % time.time()
        destFile = self.storage.getCurrentURL(testFileName)["Value"]
        fileTuple = (srcFile, destFile, fileSize)
        putFileRes = self.storage.putFile(fileTuple)
        # Check that we can get the file metadata
        getMetadataRes = self.storage.getFileMetadata(destFile)
        # Clean up the remote mess
        removeFileRes = self.storage.removeFile(destFile)
        # See what happens with non existant files
        failedMetadataRes = self.storage.getFileMetadata(destFile)
        # Check directories are handled properly
        destDir = os.path.dirname(destFile)
        directoryMetadataRes = self.storage.getFileMetadata(destDir)

        # Check the put file operation
        self.assertTrue(putFileRes["OK"])
        self.assertTrue(destFile in putFileRes["Value"]["Successful"])
        # Check the get metadata operation
        self.assertTrue(getMetadataRes["OK"])
        self.assertTrue(destFile in getMetadataRes["Value"]["Successful"])
        fileMetaData = getMetadataRes["Value"]["Successful"][destFile]
        # self.assertTrue(fileMetaData['Cached'])
        # self.assertFalse(fileMetaData['Migrated'])
        self.assertEqual(fileMetaData["Size"], fileSize)
        # check the remove file operation
        self.assertTrue(removeFileRes["OK"])
        self.assertTrue(destFile in removeFileRes["Value"]["Successful"])
        # Check the get metadata for non existant file
        self.assertTrue(failedMetadataRes["OK"])
        self.assertTrue(destFile in failedMetadataRes["Value"]["Failed"])
        expectedError = "No such file or directory"
        self.assertEqual(failedMetadataRes["Value"]["Failed"][destFile], expectedError)
        # Check that metadata operation with a directory
        self.assertTrue(directoryMetadataRes["OK"])
        self.assertTrue(destDir in directoryMetadataRes["Value"]["Failed"])
        expectedError = "RFIOStorage.getFileMetadata: Supplied path is not a file."
        self.assertEqual(directoryMetadataRes["Value"]["Failed"][destDir], expectedError)

    def test_putGetFileSize(self):
        print(
            "\n\n#########################################################"
            "################\n\n\t\t\tGet file size test\n"
        )
        # First upload a file to the storage
        srcFile = "/etc/group"
        fileSize = getSize(srcFile)
        testFileName = "testFile.%s" % time.time()
        destFile = self.storage.getCurrentURL(testFileName)["Value"]
        fileTuple = (srcFile, destFile, fileSize)
        putFileRes = self.storage.putFile(fileTuple)
        # Check that we can get the file size
        getSizeRes = self.storage.getFileSize(destFile)
        # Clean up the remote mess
        removeFileRes = self.storage.removeFile(destFile)
        # Check non existant files
        failedSizeRes = self.storage.getFileMetadata(destFile)
        # Check directories are handled properly
        destDir = os.path.dirname(destFile)
        directorySizeRes = self.storage.getFileSize(destDir)

        # Check the put file operation
        self.assertTrue(putFileRes["OK"])
        self.assertTrue(destFile in putFileRes["Value"]["Successful"])
        # Check that we got the file size correctly
        self.assertTrue(getSizeRes["OK"])
        self.assertTrue(destFile in getSizeRes["Value"]["Successful"])
        self.assertEqual(getSizeRes["Value"]["Successful"][destFile], fileSize)
        # check the remove file operation
        self.assertTrue(removeFileRes["OK"])
        self.assertTrue(destFile in removeFileRes["Value"]["Successful"])
        # Check the get size with non existant file works properly
        self.assertTrue(failedSizeRes["OK"])
        self.assertTrue(destFile in failedSizeRes["Value"]["Failed"])
        expectedError = "No such file or directory"
        self.assertEqual(failedSizeRes["Value"]["Failed"][destFile], expectedError)
        # Check that the passing a directory is handled correctly
        self.assertTrue(directorySizeRes["OK"])
        self.assertTrue(destDir in directorySizeRes["Value"]["Failed"])
        expectedError = "RFIOStorage.getFileSize: Supplied path is not a file."
        self.assertEqual(directorySizeRes["Value"]["Failed"][destDir], expectedError)

    def test_putPrestageFile(self):
        print(
            "\n\n#########################################################"
            "################\n\n\t\t\tFile prestage test\n"
        )
        # First upload a file to the storage
        srcFile = "/etc/group"
        fileSize = getSize(srcFile)
        testFileName = "testFile.%s" % time.time()
        destFile = self.storage.getCurrentURL(testFileName)["Value"]
        fileTuple = (srcFile, destFile, fileSize)
        putFileRes = self.storage.putFile(fileTuple)
        # Check that we can issue a stage request
        prestageRes = self.storage.prestageFile(destFile)
        # Clean up the remote mess
        removeFileRes = self.storage.removeFile(destFile)

        # Check the put file operation
        self.assertTrue(putFileRes["OK"])
        self.assertTrue(destFile in putFileRes["Value"]["Successful"])
        # Check the prestage file operation
        self.assertTrue(prestageRes["OK"])
        self.assertTrue(destFile in prestageRes["Value"]["Successful"])
        self.assertTrue(prestageRes["Value"]["Successful"][destFile])
        # Check the remove file operation
        self.assertTrue(removeFileRes["OK"])
        self.assertTrue(destFile in removeFileRes["Value"]["Successful"])

        # These checks are currently disabled until a bug is fixed
        # Check what happens with deleted files
        # deletedPrestageRes = self.storage.prestageFile(destFile)
        # self.assertTrue(deletedPrestageRes['OK'])
        # self.assertTrue(deletedPrestageRes['Value']['Failed'].has_key(destFile))

        # Check what happens with non-existant files #THIS IS A BUG, REPORT IR
        # testFile = "%s-THIS-IS-DEFINATELY-NOT-A-FILE" % destFile
        # nonExistantPrestageRes= self.storage.prestageFile(testFile)
        # self.assertTrue(nonExistantPrestageRes['OK'])
        # self.assertTrue(nonExistantPrestageRes['Value']['Failed'].has_key(destFile))

    def test_putFilegetTransportURL(self):
        print(
            "\n\n#########################################################" "################\n\n\t\t\tGet tURL test\n"
        )
        # First upload a file to the storage
        srcFile = "/etc/group"
        fileSize = getSize(srcFile)
        testFileName = "testFile.%s" % time.time()
        destFile = self.storage.getCurrentURL(testFileName)["Value"]
        fileTuple = (srcFile, destFile, fileSize)
        putFileRes = self.storage.putFile(fileTuple)
        # Check that we can get a turl
        getTurlRes = self.storage.getTransportURL(destFile, ["dcap", "gsidcap"])
        # Clean up the remote mess
        removeFileRes = self.storage.removeFile(destFile)
        # Try and get a turl for a non existant file
        failedGetTurlRes = self.storage.getTransportURL(destFile, ["dcap", "gsidcap"])

        # Check the put file operation
        print(putFileRes)
        self.assertTrue(putFileRes["OK"])
        self.assertTrue(destFile in putFileRes["Value"]["Successful"])
        # check the get turl operation
        print(getTurlRes, destFile)
        self.assertTrue(getTurlRes["OK"])
        self.assertTrue(destFile in getTurlRes["Value"]["Successful"])
        # check the remove file operation
        self.assertTrue(removeFileRes["OK"])
        self.assertTrue(destFile in removeFileRes["Value"]["Successful"])
        # Check the get turl with non existant file operation
        self.assertTrue(failedGetTurlRes["OK"])
        self.assertTrue(destFile in failedGetTurlRes["Value"]["Failed"])
        expectedError = "RFIOStorage.getTransportURL: File does not exist."
        self.assertEqual(failedGetTurlRes["Value"]["Failed"][destFile], expectedError)


if __name__ == "__main__":
    suite = unittest.defaultTestLoader.loadTestsFromTestCase(StoragePlugInTestCase)
    # suite = unittest.defaultTestLoader.loadTestsFromTestCase( FileTestCase )
    # suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(DirectoryTestCase))
    testResult = unittest.TextTestRunner(verbosity=2).run(suite)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import unittest
import datetime
import json
from mock import MagicMock


from DIRAC.DataManagementSystem.Agent.RequestOperations.ReplicateAndRegister import ReplicateAndRegister
from DIRAC.DataManagementSystem.Agent.RequestOperations.MoveReplica import MoveReplica
from DIRAC.RequestManagementSystem.Client.File import File
from DIRAC.RequestManagementSystem.Client.Operation import Operation
from DIRAC.RequestManagementSystem.Client.Request import Request


class ReqOpsTestCase(unittest.TestCase):
    """Base class for the clients test cases"""

    def setUp(self):
        fcMock = MagicMock()
        ftsMock = MagicMock()

        self.rr = ReplicateAndRegister()
        self.rr.fc = fcMock
        self.rr.ftsClient = ftsMock

    def tearDown(self):
        pass


#############################################################################
class MoveReplicaSuccess(ReqOpsTestCase):
    def setUp(self):
        self.op = Operation()
        self.op.Type = "MoveFile"
        self.op.SourceSE = "%s,%s" % ("sourceSE1", "sourceSE2")
        self.op.TargetSE = "%s,%s" % ("targetSE1", "targetSE2")

        self.File = File()
        self.File.LFN = "/cta/file1"
        self.File.Size = 2
        self.File.Checksum = "011300a2"
        self.File.ChecksumType = "adler32"
        self.op.addFile(self.File)

        self.req = Request()
        self.req.addOperation(self.op)
        self.mr = MoveReplica(self.op)

        self.mr.dm = MagicMock()
        self.mr.fc = MagicMock()

    # This test needs to be fixed. It currently fails because StorageElement is not mocked
    """def test__dmTransfer( self ):

    successful = {}
    for sourceSE in self.op.sourceSEList:
      successful[sourceSE] = 'dips://' + sourceSE.lower() + ':9148/DataManagement/StorageElement' + self.File.LFN

    res = {'OK': True, 'Value': {'Successful': {self.File.LFN : successful}, 'Failed': {}}}
    self.mr.dm.getActiveReplicas.return_value = res

    res = {'OK': True, 'Value': {'Successful': {self.File.LFN : {'register': 0.1228799819946289, 'replicate': 9.872732877731323}}, 'Failed': {}}}
    self.mr.dm.replicateAndRegister.return_value = res

    res = self.mr.dmTransfer( self.File )
    self.assertTrue( res['OK'] )

    self.assertEqual( self.mr.operation.__files__[0].Status, 'Waiting' )
    self.assertEqual( self.mr.operation.Status, 'Waiting' )
    self.assertEqual( self.mr.request.Status, 'Waiting' )"""

    def test__dmRemoval(self):

        res = {"OK": True, "Value": {"Successful": {self.File.LFN: {"DIRACFileCatalog": True}}, "Failed": {}}}
        self.mr.dm.removeReplica.return_value = res

        toRemoveDict = {self.File.LFN: self.File}
        targetSEs = self.op.sourceSEList

        res = self.mr.dmRemoval(toRemoveDict, targetSEs)
        self.assertTrue(res["OK"])

        resvalue = dict([(targetSE, "") for targetSE in targetSEs])
        self.assertEqual(res["Value"], {self.File.LFN: resvalue})

        self.assertEqual(self.mr.operation.__files__[0].Status, "Done")
        self.assertEqual(self.mr.operation.Status, "Done")
        self.assertEqual(self.mr.request.Status, "Done")


class MoveReplicaFailure(ReqOpsTestCase):
    def setUp(self):
        self.op = Operation()
        self.op.Type = "MoveReplica"
        self.op.SourceSE = "%s,%s" % ("sourceSE1", "sourceSE2")
        self.op.TargetSE = "%s,%s" % ("targetSE1", "targetSE2")

        self.File = File()
        self.File.LFN = "/cta/file1"
        self.File.Size = 2
        self.File.Checksum = "011300a2"
        self.File.ChecksumType = "adler32"
        self.op.addFile(self.File)

        self.req = Request()
        self.req.addOperation(self.op)
        self.mr = MoveReplica(self.op)

        self.mr.dm = MagicMock()
        self.mr.fc = MagicMock()
        self.mr.ci = MagicMock()

    def test__dmTransfer(self):

        successful = {}
        for sourceSE in self.op.sourceSEList:
            successful[sourceSE] = "dips://" + sourceSE.lower() + ":9148/DataManagement/StorageElement" + self.File.LFN

        res = {"OK": True, "Value": ({self.File.LFN: successful}, [])}
        self.mr.ci._getCatalogReplicas.return_value = res

        res = {
            "OK": True,
            "Value": {
                "MissingAllReplicas": {},
                "NoReplicas": {},
                "MissingReplica": {},
                "SomeReplicasCorrupted": {},
                "AllReplicasCorrupted": {},
            },
        }
        self.mr.ci.compareChecksum.return_value = res

        res = {"OK": True, "Value": {"Successful": {}, "Failed": {self.File.LFN: "Unable to replicate file"}}}
        self.mr.dm.replicateAndRegister.return_value = res

        res = self.mr.dmTransfer(self.File)
        self.assertFalse(res["OK"])

        self.assertEqual(self.mr.operation.__files__[0].Status, "Waiting")
        self.assertEqual(self.mr.operation.Status, "Waiting")
        self.assertEqual(self.mr.request.Status, "Waiting")

    def test__dmRemoval(self):

        res = {
            "OK": True,
            "Value": {"Successful": {}, "Failed": {self.File.LFN: "Write access not permitted for this credential"}},
        }
        self.mr.dm.removeReplica.return_value = res

        toRemoveDict = {self.File.LFN: self.File}
        targetSEs = self.op.sourceSEList

        res = self.mr.dmRemoval(toRemoveDict, targetSEs)
        self.assertTrue(res["OK"])

        resvalue = dict([(targetSE, "Write access not permitted for this credential") for targetSE in targetSEs])
        self.assertEqual(res["Value"], {self.File.LFN: resvalue})

        self.assertEqual(self.mr.operation.__files__[0].Status, "Waiting")
        self.assertEqual(self.mr.operation.Status, "Waiting")
        self.assertEqual(self.mr.request.Status, "Waiting")


class ReplicateAndRegisterSuccess(ReqOpsTestCase):
    def test__addMetadataToFiles(self):
        resMeta = {
            "OK": True,
            "Value": {
                "Failed": {},
                "Successful": {
                    "/lhcb/1.dst": {
                        "ChecksumType": "AD",
                        "Checksum": "123456",
                        "CreationDate": datetime.datetime(2013, 12, 11, 20, 20, 21),
                        "GUID": "92F9CE97-7A62-E311-8401-0025907FD430",
                        "Mode": 436,
                        "ModificationDate": datetime.datetime(2013, 12, 11, 20, 20, 21),
                        "NumberOfLinks": 1,
                        "Size": 5846023777,
                        "Status": "-",
                    },
                    "/lhcb/2.dst": {
                        "ChecksumType": "AD",
                        "Checksum": "987654",
                        "CreationDate": datetime.datetime(2013, 12, 12, 6, 26, 52),
                        "GUID": "DAE4933A-C162-E311-8A6B-003048FEAF04",
                        "Mode": 436,
                        "ModificationDate": datetime.datetime(2013, 12, 12, 6, 26, 52),
                        "NumberOfLinks": 1,
                        "Size": 5893396937,
                        "Status": "-",
                    },
                },
            },
        }

        self.rr.fc.getFileMetadata.return_value = resMeta

        file1 = File()
        file1.LFN = "/lhcb/1.dst"
        file2 = File()
        file2.LFN = "/lhcb/2.dst"

        toSchedule = {"/lhcb/1.dst": [file1, ["SE1"], ["SE2", "SE3"]], "/lhcb/2.dst": [file2, ["SE4"], ["SE5", "SE6"]]}

        res = self.rr._addMetadataToFiles(toSchedule)
        self.assertTrue(res["OK"])

        for lfn in toSchedule:
            self.assertEqual(res["Value"][lfn].LFN, lfn)
            for attr in ("GUID", "Size", "Checksum"):
                self.assertEqual(getattr(res["Value"][lfn], attr), resMeta["Value"]["Successful"][lfn][attr])
            # AD should be transformed into Adler32
            self.assertEqual(res["Value"][lfn].ChecksumType, "ADLER32")


if __name__ == "__main__":
    suite = unittest.defaultTestLoader.loadTestsFromTestCase(ReqOpsTestCase)
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(ReplicateAndRegisterSuccess))
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(MoveReplicaSuccess))
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(MoveReplicaFailure))
    testResult = unittest.TextTestRunner(verbosity=2).run(suite)



"""
It is a helper module used to create a certain plot...
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


from DIRAC import S_OK, S_ERROR

__RCSID__ = "$Id$"


def _convertToSeconds(interval):
    """
    Converts number of minutes, hours, days, weeks, months, years into seconds
    """
    # unit symbols
    units = ["s", "m", "h", "d", "w", "M", "y"]
    # this is the number of previous units in a unit
    numbers = [1, 60, 60, 24, 7, 30.0 / 7.0, 366.0 / 30.0]
    seconds = 1.0
    for unit, num in zip(units, numbers):
        seconds *= num
        if interval.endswith(unit):
            return int(float(interval[:-1]) * seconds)
    raise ValueError("Invalid time interval '%s'" % interval)


class DBUtils(object):

    """
    .. class:: DBUtils

    It implements few methods used to create the plots.

    param: list __units it is elasticsearch specific unites
    param: list __unitvalues the units in second
    param: list __esunits used to determine the buckets size

    """

    # TODO: Maybe it is better to use the same structure we have in BasePlotter

    __esbucket = {
        "1h": "1m",
        "6h": "5m",
        "12h": "10m",
        "1d": "15m",
        "2d": "30m",
        "3.5d": "1h",
        "1w": "2h",
        "2w": "4h",
        "1M": "8h",
        "2M": "12h",
        "3M": "1d",
        "6M": "2d",
        "9M": "3d",
        "1y": "4d",
        "10y": "7d",
        "100y": "1w",
    }

    def __init__(self, db, setup):
        """c'tor

        :param self: self reference
        :param db: the database module
        :param str setup: DIRAC setup
        """
        self.__db = db
        self.__setup = setup

    def getKeyValues(self, typeName, condDict):
        """
        Get all valid key values in a type
        """
        return self.__db.getKeyValues(self.__setup, typeName, condDict)

    def _retrieveBucketedData(
        self, typeName, startTime, endTime, interval, selectField, condDict=None, grouping="", metadataDict=None
    ):
        """
        It is a wrapper class...
        """
        return self.__db.retrieveBucketedData(
            typeName=typeName,
            startTime=startTime,
            endTime=endTime,
            interval=interval,
            selectField=selectField,
            condDict=condDict,
            grouping=grouping,
            metainfo=metadataDict,
        )

    def _retrieveAggregatedData(
        self, typeName, startTime, endTime, interval, selectField, condDict=None, grouping="", metadataDict=None
    ):
        """
        Retrieve data from EL
        """
        return self.__db.retrieveAggregatedData(
            typeName=typeName,
            startTime=startTime,
            endTime=endTime,
            interval=interval,
            selectField=selectField,
            condDict=condDict,
            grouping=grouping,
            metainfo=metadataDict,
        )

    def _determineBucketSize(self, start, end):
        """It is used to determine the bucket size using _esUnits

        :param int start: epoch time
        :param int end: epoch time
        :return: S_OK/S_ERROR with tuple of (binUnit, seconds)
        """
        diff = end - start

        error = "Can not determine the bucket size..."
        bucketSeconds = {}
        try:
            # Convert intervals into seconds
            for interval, binUnit in self.__esbucket.items():
                bucketSeconds[_convertToSeconds(interval)] = (binUnit, _convertToSeconds(binUnit))
            # Determine bin size according to time span
            for interval in sorted(bucketSeconds):
                if diff <= interval:
                    return S_OK(bucketSeconds[interval])
        except ValueError as e:
            error += ": " + repr(e)
        return S_ERROR(error)

    def _divideByFactor(self, dataDict, factor):
        """
        Divide by factor the values and get the maximum value
          - dataDict = { 'key' : { time1 : value,  time2 : value... }, 'key2'.. }
        """
        maxValue = 0.0
        for key in dataDict:
            currentDict = dataDict[key]
            for timeEpoch in currentDict:
                currentDict[timeEpoch] /= float(factor)
                maxValue = max(maxValue, currentDict[timeEpoch])
        return dataDict, maxValue

    def _getAccumulationMaxValue(self, dataDict):
        """
        Divide by factor the values and get the maximum value
          - dataDict = { 'key' : { time1 : value,  time2 : value... }, 'key2'.. }
        """
        maxValue = 0
        maxEpoch = 0
        for key in dataDict:
            currentDict = dataDict[key]
            for timeEpoch in currentDict:
                if timeEpoch > maxEpoch:
                    maxEpoch = timeEpoch
                    maxValue = 0
                if timeEpoch == maxEpoch:
                    maxValue += currentDict[timeEpoch]
        return maxValue



""" X509Chain is a class for managing X509 chains with their Pkeys

Link to the RFC 3820: https://tools.ietf.org/html/rfc3820
In particular, limited proxy: https://tools.ietf.org/html/rfc3820#section-3.8

There are also details available about Per-User Sub-Proxies (PUSP)
here: https://wiki.egi.eu/wiki/Usage_of_the_per_user_sub_proxy_in_EGI

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import six
import copy
import os
import stat
import tempfile
import hashlib

import re

import M2Crypto

from io import open

from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Utilities import DErrno
from DIRAC.Core.Utilities.Decorators import executeOnlyIf, deprecated
from DIRAC.ConfigurationSystem.Client.Helpers import Registry
from DIRAC.Core.Security.m2crypto import PROXY_OID, LIMITED_PROXY_OID, DIRAC_GROUP_OID
from DIRAC.Core.Security.m2crypto.X509Certificate import X509Certificate


# Decorator to check that _certList is not empty
needCertList = executeOnlyIf("_certList", S_ERROR(DErrno.ENOCHAIN))
# Decorator to check that the PKey has been loaded
needPKey = executeOnlyIf("_keyObj", S_ERROR(DErrno.ENOPKEY))


class X509Chain(object):
    """
    An X509Chain is basically a list of X509Certificate object, as well as a PKey object,
    which is associated to the X509Certificate the lowest in the chain.

    This is what you will want to use for user certificate (because they will turn into proxy....), and for
    proxy.

    A priori, once we get rid of pyGSI, we could even meld the X509Certificate into this one, and use the X509Chain
    for host certificates. After all, a certificate is nothing but a chain of length 1...

    There are normally 4 ways you would instanciate an X509Chain object:

    * You are loading a proxy from a file
    * Loading the chain from a file
    * You are getting information about your peer during an SSL connection
    * You are delegating

    Typical usages of X509Chain are illustrated bellow

    Loading a proxy from a file (this will load the chain and the key, assuming the key is in the same file)::

      proxy = X509Chain()
      res = proxy.loadProxyFromFile(myFile)
      if not res['OK']:
        return res


    Generating a proxy from a Certificate::

      cert = X509Chain()
      # Load user cert
      retVal = cert.loadChainFromFile('/home/chaen/.globus/userkey.pem')
      if not retVal['OK']:
        return retVal
      # Load the key from a different place, with a password
      retVal = cert.loadKeyFromFile('/home/chaen/.globus/userkey.pem', password='MySecretKey')
      if not retVal['OK']:
        return res

      # Generate a limited proxy, valid one hour
      retVal = cert.generateProxyToFile('/tmp/proxy.pem',
                                     3600, # only 1 h
                                     diracGroup = 'lhcb_user',
                                     strength= 2048,
                                     limited=True)


    Getting information from a peer in an SSL Connection::

      # conn is an M2Crypto.SSL.Connection instance
      chain = X509Chain.generateX509ChainFromSSLConnection(conn)
      creds = chain.getCredentials()


    Delegating a proxy to a service::

      # The server side generates a request
      # Equivalent to ProxyManager.requestDelegationUpload

      x509Req = X509Request()
      x509Req.generateProxyRequest()

      # This reqStr object is sent to the client
      reqStr = x509Req.dumpRequest()['Value']

      # This object contains both the public and private key
      pkeyReq = x509Req.getPKey()

      #######################################################

      # The client side signs the request, with its proxy
      # Assume the proxy chain was already loaded one way or the otjer

      # The proxy will contain a "bullshit private key"
      res = proxyChain.generateChainFromRequestString(reqStr, lifetime=lifetime)

      # This is sent back to the server
      delegatedProxyString = res['Value']

      ######################################################
      # Equivalent to ProxyManager.completeDelegationUpload

      # Create the new chain
      # the pkey was generated together with the Request
      delegatedProxy = X509Chain(keyObj=pkeyReq)
      delegatedProxy.loadChainFromString(delegatedProxyString)

      # make sure the public key match between Request and the new Chain
      # (Stupid, of course it will ! But it is done in the ProxyManager...)
      res = x509Req.checkChain(delegatedProxy)

    """

    def __init__(self, certList=False, keyObj=False):
        """
        C'tor

        :param certList: list of X509Certificate to constitute the chain
        :param keyObj: ~M2Crypto.EVP.PKey object. The public or public/private key associated to
                       the last certificate of the chain

        """

        # __isProxy is True if this chain represents a proxy
        self.__isProxy = False
        # Whether the proxy is limited or not
        self.__isLimitedProxy = False

        # This is the position of the first proxy in the chain
        self.__firstProxyStep = 0

        # Cache for sha1 hash of the object
        # This is just used as a unique identifier for
        # indexing in the ProxyCache
        self.__hash = False

        # List of X509Certificate constituing the chain
        # The certificate in position N has been generated from the (N+1)
        self._certList = []

        # Place holder for the EVP.PKey object
        self._keyObj = None

        if certList:
            # copy the content of the list, without copying the objects themselves
            self._certList = copy.copy(certList)
            # Immediately check if it is a proxy
            self.__checkProxyness()

        if keyObj:
            self._keyObj = keyObj

    @classmethod
    @deprecated("Use loadChainFromFile instead", onlyOnce=True)
    def instanceFromFile(cls, chainLocation):
        """Class method to generate a X509Chain from a file

        :param chainLocation: path to the file

        :returns: S_OK(X509Chain)
        """
        chain = cls()
        result = chain.loadChainFromFile(chainLocation)
        if not result["OK"]:
            return result

        return S_OK(chain)

    @staticmethod
    def generateX509ChainFromSSLConnection(sslConnection):
        """Returns an instance of X509Chain from the SSL connection

        :param sslConnection: ~M2Crypto.SSl.Connection instance

        :returns: a X509Chain instance
        """
        certList = []

        certStack = sslConnection.get_peer_cert_chain()
        for cert in certStack:
            certList.append(X509Certificate(x509Obj=cert))

        # Servers don't receive the whole chain, the last cert comes alone
        # if not self.infoDict['clientMode']:
        certList.insert(0, X509Certificate(x509Obj=sslConnection.get_peer_cert()))
        peerChain = X509Chain(certList=certList)

        return peerChain

    def loadChainFromFile(self, chainLocation):
        """
        Load a x509 chain from a pem file

        :param chainLocation: path to the file

        :returns: S_OK/S_ERROR
        """
        try:
            with open(chainLocation, "rb") as fd:
                pemData = fd.read()
        except IOError as e:
            return S_ERROR(DErrno.EOF, "%s: %s" % (chainLocation, repr(e).replace(",)", ")")))
        return self.loadChainFromString(pemData)

    def loadChainFromString(self, data):
        """
        Load a x509 cert from a string containing the pem data

        :param data: data representing the chain of certificate in the

        Return : S_OK / S_ERROR
        """
        try:
            self._certList = self.__certListFromPemString(data)
        except Exception as e:
            return S_ERROR(DErrno.ECERTREAD, "%s" % repr(e).replace(",)", ")"))

        if not self._certList:
            return S_ERROR(DErrno.EX509)

        # Update internals
        self.__checkProxyness()
        return S_OK()

    @staticmethod
    def __certListFromPemString(certString):
        """
        Create certificates list from string. String should contain certificates, just like plain text proxy file.
        """
        # To get list of X509 certificates (not X509 Certificate Chain) from string it has to be parsed like that
        # (constructors are not able to deal with big string)
        certList = []
        # If the certificate is downloaded from the server it will be a str in Python 3
        if six.PY3 and isinstance(certString, six.string_types):
            certString = certString.encode()
        pattern = r"(-----BEGIN CERTIFICATE-----((.|\n)*?)-----END CERTIFICATE-----)"
        for cert in re.findall(pattern.encode("utf-8"), certString):
            certList.append(X509Certificate(certString=cert[0]))
        return certList

    # Not used in m2crypto version
    # def setChain(self, certList):
    #   """
    #   Set the chain
    #   Return : S_OK / S_ERROR
    #   """
    #   self._certList = certList
    #   self.__loadedChain = True
    #   return S_OK()

    def loadKeyFromFile(self, chainLocation, password=False):
        """
        Load a PKey from a pem file

        :param chainLocation: path to the file
        :param password: password to decode the file.

        :returns: S_OK / S_ERROR
        """
        try:
            with open(chainLocation, "rb") as fd:
                pemData = fd.read()
        except Exception as e:
            return S_ERROR(DErrno.EOF, "%s: %s" % (chainLocation, repr(e).replace(",)", ")")))
        return self.loadKeyFromString(pemData, password)

    def loadKeyFromString(self, pemData, password=False):
        """
        Load a PKey from a string containing the pem data

        :param pemData: pem data of the key, potentially encoded with the password
        :param password: password to decode the file.

        :returns: S_OK / S_ERROR
        """
        self._keyObj = None
        if password:
            password = password.encode()
        try:
            self._keyObj = M2Crypto.EVP.load_key_string(pemData, lambda x: password)
        except Exception as e:
            return S_ERROR(DErrno.ECERTREAD, "%s (Probably bad pass phrase?)" % repr(e).replace(",)", ")"))

        return S_OK()

    def setPKey(self, pkeyObj):
        """
        Set the chain
        Return : S_OK / S_ERROR
        """
        self._keyObj = pkeyObj
        return S_OK()

    def loadProxyFromFile(self, chainLocation):
        """
        Load a Proxy from a pem file, that is both the Cert chain and the PKey

        :param chainLocation: path to the proxy file

        :returns: S_OK  / S_ERROR
        """
        try:
            with open(chainLocation, "rb") as fd:
                pemData = fd.read()
        except Exception as e:
            return S_ERROR(DErrno.EOF, "%s: %s" % (chainLocation, repr(e).replace(",)", ")")))
        return self.loadProxyFromString(pemData)

    def loadProxyFromString(self, pemData):
        """
        Load a Proxy from a pem buffer, that is both the Cert chain and the PKey

        :param pemData: PEM encoded cert chain and pkey

        :returns: S_OK / S_ERROR
        """
        retVal = self.loadChainFromString(pemData)
        if not retVal["OK"]:
            return retVal

        return self.loadKeyFromString(pemData)

    @staticmethod
    def __getProxyExtensionList(diracGroup=False, rfcLimited=False):
        """
        Get an extension stack containing the necessary extension for a proxy.
        Basically the keyUsage, the proxyCertInfo, and eventually the diracGroup

        :param diracGroup: name of the dirac group for the proxy
        :param rfcLimited: boolean to generate for a limited proxy

        :returns: M2Crypto.X509.X509_Extension_Stack object.
        """

        extStack = M2Crypto.X509.X509_Extension_Stack()

        # Standard certificate extensions
        kUext = M2Crypto.X509.new_extension(
            "keyUsage", "digitalSignature, keyEncipherment, dataEncipherment", critical=1
        )
        extStack.push(kUext)

        # Mandatory extension to be a proxy
        policyOID = LIMITED_PROXY_OID if rfcLimited else PROXY_OID
        ext = M2Crypto.X509.new_extension("proxyCertInfo", "critical, language:%s" % (policyOID), critical=1)
        extStack.push(ext)

        # Add a dirac group
        if diracGroup and isinstance(diracGroup, six.string_types):
            # the str cast is needed because M2Crypto does not play it cool with unicode here it seems
            # Also one needs to specify the ASN1 type. That's what it is...
            dGext = M2Crypto.X509.new_extension(DIRAC_GROUP_OID, str("ASN1:IA5:%s" % diracGroup))
            extStack.push(dGext)

        return extStack

    @needCertList
    def getCertInChain(self, certPos=0):
        """
        Get then a certificate in the chain

        :warning: Contrary to the pygsi version, this is not a copy!

        :param certPos: position of the certificate in the chain. Default: 0

        :returns: S_OK(X509Certificate)/S_ERROR
        """
        return S_OK(self._certList[certPos])

    @needCertList
    def getIssuerCert(self):
        """
        Returns the issuer certificate of the last one if it is a proxy, otherwise
        the last one in the chain

        :returns: S_OK(X509Certificate)/S_ERROR
        """
        if self.__isProxy:
            return S_OK(self._certList[self.__firstProxyStep + 1])
        return S_OK(self._certList[-1])

    @deprecated("Only here for compatibility reason", onlyOnce=True)
    @needPKey
    def getPKeyObj(self):
        """
        Get the pkey obj

        :returns: ~M2Crypto.EVP.PKey object
        """
        return S_OK(self._keyObj)

    @deprecated("Only here for compatibility reason")
    @needCertList
    def getCertList(self):
        """
        Get the cert list
        """
        return S_OK(self._certList)

    @needCertList
    def getNumCertsInChain(self):
        """
        length of the certificate chain

        :returns: length of the certificate chain


        """
        return S_OK(len(self._certList))

    # pylint: disable=unused-argument
    @needCertList
    @needPKey
    def generateProxyToString(self, lifetime, diracGroup=False, strength=1024, limited=False, proxyKey=False, rfc=True):
        """
        Generate a proxy and get it as a string.

        Check here: https://github.com/eventbrite/m2crypto/blob/master/demo/x509/ca.py#L45

        Args:
            lifetime (int): expected lifetime in seconds of proxy
            diracGroup (str): diracGroup to add to the certificate
            strength (int): length in bits of the pair if proxyKey not given (default 1024)
            limited (bool): Create a limited proxy (default False)
            proxyKey: M2Crypto.EVP.PKey instance with private and public key. If not given, generate one
            rfc: placeholder for backward compatibility and ignored

        :returns: S_OK(PEM encoded string), S_ERROR. The PEM string contains all the certificates in the chain
                  and the private key associated to the last X509Certificate just generated.
        """

        issuerCert = self._certList[0]

        if not proxyKey:
            # Generating key is a two step process: create key object and then assign RSA key.
            # This contains both the private and public key
            proxyKey = M2Crypto.EVP.PKey()
            proxyKey.assign_rsa(M2Crypto.RSA.gen_key(strength, 65537, callback=M2Crypto.util.quiet_genparam_callback))

        # Generate a new X509Certificate object
        proxyExtensions = self.__getProxyExtensionList(diracGroup, limited)
        res = X509Certificate.generateProxyCertFromIssuer(issuerCert, proxyExtensions, proxyKey, lifetime=lifetime)
        if not res["OK"]:
            return res
        proxyCert = res["Value"]

        # Sign it with one owns key
        proxyCert.sign(self._keyObj, "sha256")

        # Generate the proxy string
        proxyString = b"%s%s" % (
            proxyCert.asPem(),
            proxyKey.as_pem(cipher=None, callback=M2Crypto.util.no_passphrase_callback),
        )
        for i in range(len(self._certList)):
            crt = self._certList[i]
            proxyString += crt.asPem()
        return S_OK(proxyString)

    # pylint: disable=unused-argument
    def generateProxyToFile(self, filePath, lifetime, diracGroup=False, strength=1024, limited=False, rfc=True):
        """
        Generate a proxy and put it into a file

        Args:
            filePath: file to write
            lifetime: expected lifetime in seconds of proxy
            diracGroup: diracGroup to add to the certificate
            strength: length in bits of the pair
            limited: Create a limited proxy
            rfc: placeholder and ignored
        """
        retVal = self.generateProxyToString(lifetime, diracGroup, strength, limited)
        if not retVal["OK"]:
            return retVal
        try:
            with open(filePath, "wb") as fd:
                fd.write(retVal["Value"])
        except Exception as e:
            return S_ERROR(DErrno.EWF, "%s :%s" % (filePath, repr(e).replace(",)", ")")))
        try:
            os.chmod(filePath, stat.S_IRUSR | stat.S_IWUSR)
        except Exception as e:
            return S_ERROR(DErrno.ESPF, "%s :%s" % (filePath, repr(e).replace(",)", ")")))
        return S_OK()

    @needCertList
    def isProxy(self):
        """
         Check whether this chain is a proxy

        :returns: S_OK(boolean)
        """
        return S_OK(self.__isProxy)

    @needCertList
    def isLimitedProxy(self):
        """
        Check whether this chain is a limited proxy

        :returns: S_OK(boolean)
        """
        return S_OK(self.__isProxy and self.__isLimitedProxy)

    @needCertList
    def isValidProxy(self, ignoreDefault=False):
        """
        Check whether this chain is a valid proxy, that is:
          * a proxy
          * still valid
          * with a valid group

        :param ignoreDefault: (what a stupid name) if True, do not lookup the CS

        :returns: S_OK(True) if the proxy is valid, S_ERROR otherwise

        """
        if not self.__isProxy:
            return S_ERROR(DErrno.ENOCHAIN, "Chain is not a proxy")

        if self.hasExpired()["Value"]:
            return S_ERROR(DErrno.ENOCHAIN)

        if ignoreDefault:
            groupRes = self.getDIRACGroup(ignoreDefault=True)
            if not groupRes["OK"]:
                return groupRes
            if not groupRes["Value"]:
                return S_ERROR(DErrno.ENOGROUP)

        return S_OK(True)

    def isVOMS(self):
        """
        Check whether this proxy contains VOMS extensions.
        It is enough for one of the certificate of the chain to have VOMS extension

        :returns: S_OK(boolean)
        """

        if not self.__isProxy:
            return S_OK(False)

        for cert in self._certList:
            if cert.hasVOMSExtensions()["Value"]:
                return S_OK(True)
        return S_OK(False)

    def isRFC(self):
        """Check whether this is an RFC proxy. It can only be true, providing it is a proxy

        :returns: S_OK(boolean)
        """

        return self.isProxy()

    def getVOMSData(self):
        """
        Returns the voms data.

        :returns: See :py:func:`~DIRAC.Core.Security.m2crypto.X509Certificate.getVOMSData`
                  If no VOMS data is available, return DErrno.EVOMS
        :warning: In case the chain is not a proxy, this method will return False.
                  Yes, it's stupid, but it is for compatibility...

        """
        if not self.__isProxy:
            return S_OK(False)

        for cert in self._certList:
            res = cert.getVOMSData()
            if res["OK"]:
                return res
        return S_ERROR(DErrno.EVOMS)

    def __checkProxyness(self):
        """This method is called upon initialization of a chain and fill in some internal attributes.

        Pure madness...

        To me, this method just seems to work by pure luck..
        """

        self.__hash = False
        self.__firstProxyStep = len(self._certList) - 2  # -1 is user cert by default, -2 is first proxy step
        self.__isProxy = True
        self.__isLimitedProxy = False
        prevDNMatch = 2
        # If less than 2 steps in the chain is no proxy
        if len(self._certList) < 2:
            self.__isProxy = False
            return

        # Here we make sure that each certificate in the chain was
        # signed by the previous one
        for step in range(len(self._certList) - 1):

            # this is a cryptographic check with the keys
            issuerMatch = self.__checkIssuer(step, step + 1)
            if not issuerMatch:
                self.__isProxy = False
                return

            # Do we need to check the proxy DN?
            if prevDNMatch:
                dnMatch = self.__checkProxyDN(step, step + 1)
                if dnMatch == 0:
                    # If we are not in the first step we've found the entity cert
                    if step > 0:
                        self.__firstProxyStep = step - 1
                    # If we are in the first step this is not a proxy
                    else:
                        self.__isProxy = False
                        return
                # Limited proxy DN match
                elif dnMatch == 2:
                    self.__isLimitedProxy = True
                    if prevDNMatch != 2:
                        self.__isProxy = False
                        self.__isLimitedProxy = False
                        return
                prevDNMatch = dnMatch

    def __checkProxyDN(self, certStep, issuerStep):
        """
        Checks that the subject of the proxy is properly derived from the issuer subject.

        Args:
            certStep: position of the certificate to check in self.__certList
            issuerStep: position of the issuer certificate to check in self.__certList

        :returns: an int based on the match:
                  0 = no match
                  1 = proxy match
                  2 = limited proxy match
        """

        issuerSubject = self._certList[issuerStep].getSubjectNameObject()
        if not issuerSubject["OK"]:
            return 0
        issuerSubject = issuerSubject["Value"]

        proxySubject = self._certList[certStep].getSubjectNameObject()
        if not proxySubject["OK"]:
            return 0
        proxySubject = proxySubject["Value"]

        lastEntry = str(proxySubject).split("/")[-1].split("=")
        limited = False
        if lastEntry[0] != "CN":
            return 0

        # For non-RFC proxy, the proxy always had these two strings in the CN
        if lastEntry[1] not in ("proxy", "limited proxy"):
            # for RFC proxy, one has to check the extension.
            ext = self._certList[certStep].getExtension("proxyCertInfo")
            if not ext["OK"]:
                return 0

            ext = ext["Value"]

            # Check the RFC
            contraint = [
                line.split(":")[1].strip()
                for line in ext.get_value().split("\n")
                if line.split(":")[0] == "Policy Language"
            ]
            if not contraint:
                return 0
            if contraint[0] == LIMITED_PROXY_OID:
                limited = True
        else:
            if lastEntry[1] == "limited proxy":
                limited = True
        if not str(issuerSubject) == str(proxySubject)[: str(proxySubject).rfind("/")]:
            return 0
        return 1 if not limited else 2

    def __checkIssuer(self, certStep, issuerStep):
        """
        Check that the issuer has signed the certificate with his private key

        :param certStep: position of the certificate in self.__certList
        :param issuerStep: position of the issuer certificate

        :returns: S_OK(boolean)

        """
        issuerCert = self._certList[issuerStep]
        cert = self._certList[certStep]
        pubKey = issuerCert.getPublicKey()["Value"]

        return cert.verify(pubKey)["Value"]

    @needCertList
    def getDIRACGroup(self, ignoreDefault=False):
        """
        Retrieve the dirac group of the chain

        :param ignoreDefault: (default False) if True, do not lookup the CS for a group if it is not in the proxy

        :returns: S_OK(dirac group)/S_ERROR
        """
        if not self.__isProxy:
            return S_ERROR(DErrno.EX509, "Chain does not contain a valid proxy")

        # If it is a PUSP, we do a lookup based on the certificate
        # (you can't do a PUSP out of a proxy)
        if self.isPUSP()["Value"]:
            return self._certList[self.__firstProxyStep - 2].getDIRACGroup(ignoreDefault=ignoreDefault)

        # The code below will find the first match of the DIRAC group
        for cert in reversed(self._certList):
            # We specifically say we do not want the default to first check inside the proxy
            retVal = cert.getDIRACGroup(ignoreDefault=True)
            if retVal["OK"] and "Value" in retVal and retVal["Value"]:
                return retVal

        # No DIRAC group found, try to get the default one
        return self.getCertInChain(self.__firstProxyStep)["Value"].getDIRACGroup(ignoreDefault=ignoreDefault)

    @needCertList
    def hasExpired(self):
        """
        Check whether any element of the chain has expired

        :returns: S_OK(boolean)
        """
        for cert in reversed(self._certList):
            res = cert.hasExpired()
            if not res["OK"]:
                return res
            # If True, it means the cert has expired
            if res["Value"]:
                return S_OK(True)

        return S_OK(False)

    @needCertList
    def getNotAfterDate(self):
        """
        Get the smallest not after date

        :returns: S_OK(datetime.datetime)
        """
        notAfter = self._certList[0].getNotAfterDate()
        if not notAfter["OK"]:
            return notAfter
        notAfter = notAfter["Value"]

        for cert in reversed(self._certList):
            res = cert.getNotAfterDate()
            if not res["OK"]:
                return res
            stepNotAfter = res["Value"]

            # If the current cert has already expired
            # we return this as notAfter date
            res = cert.hasExpired()
            if not res["OK"]:
                return res
            if res["Value"]:
                return S_OK(stepNotAfter)

            # if the current cert has a shorter lifetime
            # as the current reference, take it as new reference
            notAfter = min(notAfter, stepNotAfter)

        return S_OK(notAfter)

    @needCertList
    def generateProxyRequest(self, bitStrength=1024, limited=False):
        """
        Generate a proxy request.
        See :py:meth:`DIRAC.Core.Security.m2crypto.X509Certificate.X509Certificate.generateProxyRequest`

        Return S_OK( X509Request ) / S_ERROR
        """

        # We use the first certificate of the chain to do the proxy request
        x509 = self._certList[0]
        return x509.generateProxyRequest(bitStrength, limited)

    @needCertList
    @needPKey
    def generateChainFromRequestString(self, pemData, lifetime=86400, requireLimited=False, diracGroup=False, rfc=True):
        """
        Generate a x509 chain from a request.

        :param pemData: PEM encoded request
        :param lifetime: lifetime of the delegated proxy in seconds (default 1 day)
        :param requireLimited: if True, requires a limited proxy
        :param diracGroup: DIRAC group to put in the proxy
        :param rfc: placeholder for compatibility, ignored

        :returns: S_OK( X509 chain pem encoded string ) / S_ERROR. The new chain will have been signed
                  with the public key included in the request

        """
        try:
            req = M2Crypto.X509.load_request_string(pemData, format=M2Crypto.X509.FORMAT_PEM)

        except Exception as e:
            return S_ERROR(DErrno.ECERTREAD, "Can't load request data: %s" % repr(e).replace(",)", ")"))

        # I am not sure this test makes sense.
        # You can't request a limit proxy if you are yourself not limited ?!
        # I think it should be a "or" instead of "and"
        limited = requireLimited and self.isLimitedProxy().get("Value", False)

        return self.generateProxyToString(lifetime, diracGroup, 1024, limited, req.get_pubkey())

    @needCertList
    def getRemainingSecs(self):
        """
        Get remaining time (minimum of all cert in the chain)

        :returns: S_OK(time left in seconds)
        """
        remainingSecs = self.getCertInChain(0)["Value"].getRemainingSecs()["Value"]
        for cert in self._certList[1:]:
            stepRS = cert.getRemainingSecs()["Value"]
            remainingSecs = min(remainingSecs, stepRS)

        return S_OK(remainingSecs)

    @needCertList
    def dumpAllToString(self):
        """
        Dump the current chain as a PEM encoded string
        The order would be:

          * first certificate
          * private key (without passphrase)
          * other certificates

        :returns: S_OK(PEM encoded chain with private key)
        """
        data = self._certList[0].asPem()
        if self._keyObj:
            data += self._keyObj.as_pem(cipher=None, callback=M2Crypto.util.no_passphrase_callback)
        for cert in self._certList[1:]:
            data += cert.asPem()
        return S_OK(data)

    def dumpAllToFile(self, filename=False):
        """
        Dump all to file.

        :param filename: If not specified, a temporary one will be created

        :returns: S_OK(filename)/S_ERROR
        """
        retVal = self.dumpAllToString()
        if not retVal["OK"]:
            return retVal
        pemData = retVal["Value"]
        try:
            if not filename:
                fd, filename = tempfile.mkstemp()
                os.close(fd)
            with open(filename, "wb") as fp:
                fp.write(pemData)
        except Exception as e:
            return S_ERROR(DErrno.EWF, "%s :%s" % (filename, repr(e).replace(",)", ")")))
        try:
            os.chmod(filename, stat.S_IRUSR | stat.S_IWUSR)
        except Exception as e:
            return S_ERROR(DErrno.ESPF, "%s :%s" % (filename, repr(e).replace(",)", ")")))
        return S_OK(filename)

    @needCertList
    def dumpChainToString(self):
        """
        Dump only cert chain to string, without the PKey

        :returns: S_OK(pem chain)
        """
        return S_OK(b"".join(cert.asPem() for cert in self._certList))

    @needPKey
    def dumpPKeyToString(self):
        """
        Dump only the key to string, not encoded

        :returns: S_OK(PEM encoded key)

        """
        return S_OK(self._keyObj.as_pem(cipher=None, callback=M2Crypto.util.no_passphrase_callback))

    def __str__(self):
        """String representation"""
        repStr = "<X509Chain"
        if self._certList:
            repStr += " %s certs " % len(self._certList)
            for cert in self._certList:
                repStr += "[%s]" % str(cert.getSubjectDN()["Value"])
        if self._keyObj:
            repStr += " with key"
        repStr += ">"
        return repStr

    def __repr__(self):
        """Object representation"""
        return self.__str__()

    def isPUSP(self):
        """Checks whether the current chain is a PUSP

        :returns: S_OK(boolean).
                  If True, the S_OK structure is enriched with:
                  * Indentity: the DN
                  * SubProxyUser: name of the user
        """
        if self.__isProxy:
            # Check if we have a subproxy
            dn = self._certList[self.__firstProxyStep].getSubjectDN()
            if not dn["OK"]:
                return dn
            dn = dn["Value"]

            subproxyUser = isPUSPdn(dn)
            if subproxyUser:
                result = S_OK(True)
                result["Identity"] = dn
                result["SubproxyUser"] = subproxyUser
                return result

        return S_OK(False)

    @needCertList
    def getCredentials(self, ignoreDefault=False, withRegistryInfo=True):
        """Returns a summary of the credentials contained in the current chain

        :params ignoreDefault: (default False) If True and if no DIRAC group is found in the proxy, lookup the CS
        :params withRegistryInfo: (default True) if set to True, will enhance the returned dict with info
                                  from the registry


        :returns: S_OK with the credential dict. Some parameters of the dict are always there, other depends
                on the nature of the Chain

                Always present:
                  * subject: str. The last DN in the chain
                  * issuer: str. The issuer of the last cert in the chain
                  * secondsLeft: validity of the chain in seconds (see :py:meth:`.getRemainingSecs`)
                  * isProxy: boolean (see :py:meth:`.isProxy`)
                  * isLimitedProxy: boolean (see :py:meth:`.isLimitedProxy`)
                  * validDN: boolean if the DN is known to DIRAC
                  * validGroup: False (see further definition)
                  * DN: either the DN of the host, or the DN of the user corresponding to the proxy


                Only for proxy:
                  * identity: If it is a normal proxy, it is the DN of the certificate.
                              If it is a PUSP, it contains the identity as in :py:meth:`.isPUSP`
                  * username: DIRAC username associated to the DN (needs withRegistryInfo)
                              (see :py:func:`DIRAC.ConfigurationSystem.Client.Helpers.Registry.getUsernameForDN`)
                  * group: DIRAC group, depending on ignoreDefault param(see :py:meth:`.getDIRACGroup`)
                  * validGroup: True if the group found is in the list of groups the user belongs to
                  * groupProperty: (only if validGroup) get the properties of the group

                For Host certificate (needs withRegistryInfo):
                  * group: always `hosts`
                  * hostname: name of the host as registered in the CS
                             (see :py:func:`DIRAC.ConfigurationSystem.Client.Helpers.Registry.getHostnameForDN`)
                  * validGroup: True
                  * groupProperties: host options
                                    (see :py:func:`DIRAC.ConfigurationSystem.Client.Helpers.Registry.getHostOption`)

                If it is a user certificate (needs withRegistryInfo):
                  * username: like for proxy
                  * validDN: like proxy
        """
        credDict = {
            "subject": str(self._certList[0].getSubjectDN()["Value"]),  # ['Value'] :(
            "issuer": self._certList[0].getIssuerDN()["Value"],  # ['Value'] :(
            "secondsLeft": self.getRemainingSecs()["Value"],
            "isProxy": self.__isProxy,
            "isLimitedProxy": self.__isProxy and self.__isLimitedProxy,
            "validDN": False,
            "validGroup": False,
        }

        # Add the DN entry as the subject.
        credDict["DN"] = credDict["subject"]
        if self.__isProxy:
            credDict["identity"] = str(
                self._certList[self.__firstProxyStep + 1].getSubjectDN()["Value"]
            )  # ['Value'] :(
            # if the chain is a proxy, then the DN we want to work with is the real one of the
            # user, not the one of his proxy
            credDict["DN"] = credDict["identity"]

            # Check if we have the PUSP case
            result = self.isPUSP()
            if result["OK"] and result["Value"]:
                credDict["identity"] = result["Identity"]
                credDict["subproxyUser"] = result["SubproxyUser"]

            if withRegistryInfo:
                retVal = Registry.getUsernameForDN(credDict["identity"])
                if not retVal["OK"]:
                    return S_OK(credDict)
                credDict["username"] = retVal["Value"]
                credDict["validDN"] = True
            retVal = self.getDIRACGroup(ignoreDefault=ignoreDefault)
            if retVal["OK"]:
                diracGroup = retVal["Value"]
                credDict["group"] = diracGroup
                if withRegistryInfo:
                    retVal = Registry.getGroupsForUser(credDict["username"])
                    if retVal["OK"] and diracGroup in retVal["Value"]:
                        credDict["validGroup"] = True
                        credDict["groupProperties"] = Registry.getPropertiesForGroup(diracGroup)
        elif withRegistryInfo:
            retVal = Registry.getHostnameForDN(credDict["subject"])
            if retVal["OK"]:
                credDict["group"] = "hosts"
                credDict["hostname"] = retVal["Value"]
                credDict["validDN"] = True
                credDict["validGroup"] = True
                credDict["groupProperties"] = Registry.getHostOption(credDict["hostname"], "Properties")

            retVal = Registry.getUsernameForDN(credDict["subject"])
            if retVal["OK"]:
                credDict["username"] = retVal["Value"]
                credDict["validDN"] = True
        return S_OK(credDict)

    @needCertList
    def hash(self):
        """Get a hash of the chain
        In practice, this is only used to index the chain in a DictCache

        :returns: S_OK(string hash)
        """
        if self.__hash:
            return S_OK(self.__hash)
        sha1 = hashlib.sha1()
        for cert in self._certList:
            sha1.update(str(cert.getSubjectNameObject()["Value"]).encode())
        sha1.update(str(self.getRemainingSecs()["Value"] / 3600).encode())
        sha1.update(self.getDIRACGroup()["Value"].encode())
        if self.isVOMS():
            sha1.update(b"VOMS")
            from DIRAC.Core.Security.VOMS import VOMS

            result = VOMS().getVOMSAttributes(self)
            if result["OK"]:
                for attribute in result["Value"]:
                    sha1.update(attribute.encode())
        self.__hash = sha1.hexdigest()
        return S_OK(self.__hash)


def isPUSPdn(userDN):
    """Evaluate if the DN is of the PUSP type or not

    :param str userDN: user DN string

    :returns: the subproxy user name or None
    """
    lastEntry = userDN.split("/")[-1].split("=")
    if lastEntry[0] == "CN" and lastEntry[1].startswith("user:"):
        return userDN.split("/")[-1].split(":")[1]
    return None



"""
Provides Accounting functionality to DIRAC
It includes 2 different Services:

* DataStore: where new records are inserted
* ReportGenerator: that produce reports using the inserted records

and the associated Clients:
* DataStoreClient
* ReportsClient

DIRAC Accounting uses a number of predefined Types that must include:
* Accounting keys (text) to classify the records
* Accounting fields (numeric) to included the accounted data
* bucket definition to set the granularity of the reports

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__package__ = "DIRAC.AccountingSystem"
